"""
This script creates a test that fails when garage.tf.algos.TRPO performance is
too low.
"""
import gym
import pytest

from garage.envs import normalize
from garage.experiment import snapshotter
from garage.np.baselines import LinearFeatureBaseline
from garage.tf.algos import TRPO
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.optimizers import FiniteDifferenceHvp
from garage.tf.policies import CategoricalGRUPolicyWithModel
from garage.tf.policies import CategoricalLSTMPolicyWithModel
from tests.fixtures import snapshot_config, TfGraphTestCase


class TestTRPO(TfGraphTestCase):

    @pytest.mark.large
    def test_trpo_lstm_cartpole(self):
        with LocalTFRunner(snapshot_config, sess=self.sess) as runner:
            env = TfEnv(normalize(gym.make('CartPole-v1')))

            policy = CategoricalLSTMPolicyWithModel(name='policy',
                                                    env_spec=env.spec)

            baseline = LinearFeatureBaseline(env_spec=env.spec)

            algo = TRPO(env_spec=env.spec,
                        policy=policy,
                        baseline=baseline,
                        max_path_length=100,
                        discount=0.99,
                        max_kl_step=0.01,
                        optimizer_args=dict(hvp_approach=FiniteDifferenceHvp(
                            base_eps=1e-5)))

            snapshotter.snapshot_dir = './'
            runner.setup(algo, env)
            last_avg_ret = runner.train(n_epochs=10, batch_size=2048)
            assert last_avg_ret > 80

            env.close()

    @pytest.mark.large
    def test_trpo_gru_cartpole(self):
        with LocalTFRunner(snapshot_config, sess=self.sess) as runner:
            env = TfEnv(normalize(gym.make('CartPole-v1')))

            policy = CategoricalGRUPolicyWithModel(name='policy',
                                                   env_spec=env.spec)

            baseline = LinearFeatureBaseline(env_spec=env.spec)

            algo = TRPO(env_spec=env.spec,
                        policy=policy,
                        baseline=baseline,
                        max_path_length=100,
                        discount=0.99,
                        max_kl_step=0.01,
                        optimizer_args=dict(hvp_approach=FiniteDifferenceHvp(
                            base_eps=1e-5)))

            runner.setup(algo, env)
            last_avg_ret = runner.train(n_epochs=10, batch_size=2048)
            assert last_avg_ret > 80

            env.close()



from garage.tf.q_functions.base import QFunction
from garage.tf.q_functions.base2 import QFunction2
from garage.tf.q_functions.continuous_mlp_q_function import (
    ContinuousMLPQFunction)
from garage.tf.q_functions.continuous_mlp_q_function_with_model import (
    ContinuousMLPQFunctionWithModel)
from garage.tf.q_functions.discrete_cnn_q_function import DiscreteCNNQFunction
from garage.tf.q_functions.discrete_mlp_q_function import DiscreteMLPQFunction

__all__ = [
    'QFunction', 'QFunction2', 'ContinuousMLPQFunction',
    'ContinuousMLPQFunctionWithModel', 'DiscreteCNNQFunction',
    'DiscreteMLPQFunction'
]



"""
Unit test for GaussianLSTMPolicy with Model.

This test consists of four different GaussianLSTMPolicy: P1, P2, P3
and P4. P1 and P2 are from GaussianLSTMPolicy, which does not use
garage.tf.models.LSTMModel while P3 and P4 do use.

This test ensures the outputs from all the policies are the same,
for the transition from using GaussianLSTMPolicy to
GaussianLSTMPolicyWithModel.

It covers get_action, get_actions, dist_info_sym, kl_sym,
log_likelihood_sym, entropy_sym and likelihood_ratio_sym.
"""
from unittest import mock

import numpy as np
import tensorflow as tf

from garage.tf.envs import TfEnv
from garage.tf.misc import tensor_utils
from garage.tf.policies import GaussianLSTMPolicy
from garage.tf.policies import GaussianLSTMPolicyWithModel
from tests.fixtures import TfGraphTestCase
from tests.fixtures.envs.dummy import DummyBoxEnv


class TestGaussianLSTMPolicyWithModelTransit(TfGraphTestCase):
    def setup_method(self):
        super().setup_method()
        with mock.patch('tensorflow.random.normal') as mock_rand:
            mock_rand.return_value = 0.5
            env = TfEnv(DummyBoxEnv(obs_dim=(1, ), action_dim=(1, )))
            self.default_initializer = tf.constant_initializer(1)
            self.default_hidden_nonlinearity = tf.nn.tanh
            self.default_recurrent_nonlinearity = tf.nn.sigmoid
            self.default_output_nonlinearity = None
            self.time_step = 1

            self.policy1 = GaussianLSTMPolicy(
                env_spec=env.spec,
                hidden_dim=4,
                hidden_nonlinearity=self.default_hidden_nonlinearity,
                recurrent_nonlinearity=self.default_recurrent_nonlinearity,
                recurrent_w_x_init=self.default_initializer,
                recurrent_w_h_init=self.default_initializer,
                output_nonlinearity=self.default_output_nonlinearity,
                output_w_init=self.default_initializer,
                state_include_action=True,
                name='P1')
            self.policy2 = GaussianLSTMPolicy(
                env_spec=env.spec,
                hidden_dim=4,
                hidden_nonlinearity=self.default_hidden_nonlinearity,
                recurrent_nonlinearity=self.default_recurrent_nonlinearity,
                recurrent_w_x_init=self.default_initializer,
                recurrent_w_h_init=self.default_initializer,
                output_nonlinearity=self.default_output_nonlinearity,
                output_w_init=tf.constant_initializer(2),
                state_include_action=True,
                name='P2')

            self.sess.run(tf.compat.v1.global_variables_initializer())

            self.policy3 = GaussianLSTMPolicyWithModel(
                env_spec=env.spec,
                hidden_dim=4,
                hidden_nonlinearity=self.default_hidden_nonlinearity,
                hidden_w_init=self.default_initializer,
                recurrent_nonlinearity=self.default_recurrent_nonlinearity,
                recurrent_w_init=self.default_initializer,
                output_nonlinearity=self.default_output_nonlinearity,
                output_w_init=self.default_initializer,
                state_include_action=True,
                name='P3')
            self.policy4 = GaussianLSTMPolicyWithModel(
                env_spec=env.spec,
                hidden_dim=4,
                hidden_nonlinearity=self.default_hidden_nonlinearity,
                hidden_w_init=self.default_initializer,
                recurrent_nonlinearity=self.default_recurrent_nonlinearity,
                recurrent_w_init=self.default_initializer,
                output_nonlinearity=self.default_output_nonlinearity,
                output_w_init=tf.constant_initializer(2),
                state_include_action=True,
                name='P4')

            self.policy1.reset()
            self.policy2.reset()
            self.policy3.reset()
            self.policy4.reset()
            self.obs = [env.reset()]
            self.obs = np.concatenate(
                [self.obs for _ in range(self.time_step)], axis=0)

            self.obs_ph = tf.compat.v1.placeholder(
                tf.float32, shape=(None, None, env.observation_space.flat_dim))
            self.action_ph = tf.compat.v1.placeholder(
                tf.float32, shape=(None, None, env.action_space.flat_dim))

            self.dist1_sym = self.policy1.dist_info_sym(
                obs_var=self.obs_ph,
                state_info_vars={
                    'prev_action': np.zeros((2, self.time_step, 1))
                },
                name='p1_sym')
            self.dist2_sym = self.policy2.dist_info_sym(
                obs_var=self.obs_ph,
                state_info_vars={
                    'prev_action': np.zeros((2, self.time_step, 1))
                },
                name='p2_sym')
            self.dist3_sym = self.policy3.dist_info_sym(
                obs_var=self.obs_ph,
                state_info_vars={
                    'prev_action': np.zeros((2, self.time_step, 1))
                },
                name='p3_sym')
            self.dist4_sym = self.policy4.dist_info_sym(
                obs_var=self.obs_ph,
                state_info_vars={
                    'prev_action': np.zeros((2, self.time_step, 1))
                },
                name='p4_sym')

    def test_dist_info_sym_output(self):
        # batch size = 2
        dist1 = self.sess.run(
            self.dist1_sym, feed_dict={self.obs_ph: [self.obs, self.obs]})
        dist2 = self.sess.run(
            self.dist2_sym, feed_dict={self.obs_ph: [self.obs, self.obs]})
        dist3 = self.sess.run(
            self.dist3_sym, feed_dict={self.obs_ph: [self.obs, self.obs]})
        dist4 = self.sess.run(
            self.dist4_sym, feed_dict={self.obs_ph: [self.obs, self.obs]})

        assert np.array_equal(dist1['mean'], dist3['mean'])
        assert np.array_equal(dist1['log_std'], dist3['log_std'])
        assert np.array_equal(dist2['mean'], dist4['mean'])
        assert np.array_equal(dist2['log_std'], dist4['log_std'])

    @mock.patch('numpy.random.normal')
    def test_get_action(self, mock_rand):
        mock_rand.return_value = 0.5

        action1, agent_info1 = self.policy1.get_action(self.obs)
        action2, agent_info2 = self.policy2.get_action(self.obs)
        action3, agent_info3 = self.policy3.get_action(self.obs)
        action4, agent_info4 = self.policy4.get_action(self.obs)

        assert np.array_equal(action1, action3)
        assert np.array_equal(action2, action4)
        assert np.array_equal(agent_info1['mean'], agent_info3['mean'])
        assert np.array_equal(agent_info1['log_std'], agent_info3['log_std'])
        assert np.array_equal(agent_info2['mean'], agent_info4['mean'])
        assert np.array_equal(agent_info2['log_std'], agent_info4['log_std'])

        actions1, agent_infos1 = self.policy1.get_actions([self.obs])
        actions2, agent_infos2 = self.policy2.get_actions([self.obs])
        actions3, agent_infos3 = self.policy3.get_actions([self.obs])
        actions4, agent_infos4 = self.policy4.get_actions([self.obs])

        assert np.array_equal(actions1, actions3)
        assert np.array_equal(actions2, actions4)
        assert np.array_equal(agent_infos1['mean'], agent_infos3['mean'])
        assert np.array_equal(agent_infos1['log_std'], agent_infos3['log_std'])
        assert np.array_equal(agent_infos2['mean'], agent_infos4['mean'])
        assert np.array_equal(agent_infos2['log_std'], agent_infos4['log_std'])

    def test_kl_sym(self):
        kl_diff_sym1 = self.policy1.distribution.kl_sym(
            self.dist1_sym, self.dist2_sym)
        objective1 = tf.reduce_mean(kl_diff_sym1)

        kl_func = tensor_utils.compile_function([self.obs_ph], objective1)
        kl1 = kl_func([self.obs, self.obs])

        kl_diff_sym2 = self.policy3.distribution.kl_sym(
            self.dist3_sym, self.dist4_sym)
        objective2 = tf.reduce_mean(kl_diff_sym2)

        kl_func = tensor_utils.compile_function([self.obs_ph], objective2)
        kl2 = kl_func([self.obs, self.obs])

        assert np.array_equal(kl1, kl2)

    def test_log_likehihood_sym(self):
        log_prob_sym1 = self.policy1.distribution.log_likelihood_sym(
            self.action_ph, self.dist1_sym)
        log_prob_func = tensor_utils.compile_function(
            [self.obs_ph, self.action_ph], log_prob_sym1)
        log_prob1 = log_prob_func([self.obs, self.obs],
                                  np.ones((2, self.time_step, 1)))

        log_prob_sym2 = self.policy3.distribution.log_likelihood_sym(
            self.action_ph, self.dist3_sym)
        log_prob_func2 = tensor_utils.compile_function(
            [self.obs_ph, self.action_ph], log_prob_sym2)
        log_prob2 = log_prob_func2([self.obs, self.obs],
                                   np.ones((2, self.time_step, 1)))
        assert np.array_equal(log_prob1, log_prob2)

        log_prob_sym1 = self.policy2.distribution.log_likelihood_sym(
            self.action_ph, self.dist2_sym)
        log_prob_func = tensor_utils.compile_function(
            [self.obs_ph, self.action_ph], log_prob_sym1)
        log_prob1 = log_prob_func([self.obs, self.obs],
                                  np.ones((2, self.time_step, 1)))

        log_prob_sym2 = self.policy4.distribution.log_likelihood_sym(
            self.action_ph, self.dist4_sym)
        log_prob_func2 = tensor_utils.compile_function(
            [self.obs_ph, self.action_ph], log_prob_sym2)
        log_prob2 = log_prob_func2([self.obs, self.obs],
                                   np.ones((2, self.time_step, 1)))
        assert np.array_equal(log_prob1, log_prob2)

    def test_policy_entropy_sym(self):
        entropy_sym1 = self.policy1.distribution.entropy_sym(
            self.dist1_sym, name='entropy_sym1')
        entropy_func = tensor_utils.compile_function([self.obs_ph],
                                                     entropy_sym1)
        entropy1 = entropy_func([self.obs, self.obs])

        entropy_sym2 = self.policy3.distribution.entropy_sym(
            self.dist3_sym, name='entropy_sym1')
        entropy_func = tensor_utils.compile_function([self.obs_ph],
                                                     entropy_sym2)
        entropy2 = entropy_func([self.obs, self.obs])
        assert np.array_equal(entropy1, entropy2)

    def test_likelihood_ratio_sym(self):
        likelihood_ratio_sym1 = self.policy1.distribution.likelihood_ratio_sym(
            self.action_ph,
            self.dist1_sym,
            self.dist2_sym,
            name='li_ratio_sym1')
        likelihood_ratio_func = tensor_utils.compile_function(
            [self.action_ph, self.obs_ph], likelihood_ratio_sym1)
        likelihood_ratio1 = likelihood_ratio_func(
            np.ones((2, 1, 1)), [self.obs, self.obs])

        likelihood_ratio_sym2 = self.policy3.distribution.likelihood_ratio_sym(
            self.action_ph,
            self.dist3_sym,
            self.dist4_sym,
            name='li_ratio_sym2')
        likelihood_ratio_func = tensor_utils.compile_function(
            [self.action_ph, self.obs_ph], likelihood_ratio_sym2)
        likelihood_ratio2 = likelihood_ratio_func(
            np.ones((2, 1, 1)), [self.obs, self.obs])

        assert np.array_equal(likelihood_ratio1, likelihood_ratio2)



"""
This script creates a test that tests functions in garage.misc.tensor_utils.
"""

import numpy as np

from garage.misc.tensor_utils import normalize_pixel_batch
from garage.tf.envs import TfEnv
from tests.fixtures.envs.dummy import DummyBoxEnv
from tests.fixtures.envs.dummy import DummyDiscretePixelEnv


class TestTensorUtil:
    def test_normalize_pixel_patch(self):
        env = TfEnv(DummyDiscretePixelEnv())
        obs = env.reset()
        obs_normalized = normalize_pixel_batch(env, obs)
        expected = [ob / 255.0 for ob in obs]
        assert np.allclose(obs_normalized, expected)

    def test_normalize_pixel_patch_not_trigger(self):
        env = TfEnv(DummyBoxEnv())
        obs = env.reset()
        obs_normalized = normalize_pixel_batch(env, obs)
        assert np.array_equal(obs, obs_normalized)



"""
This script creates a test that fails when garage.tf.algos.PPO performance is
too low.
"""
import gym
import pytest
import tensorflow as tf

from garage.envs import normalize
from garage.tf.algos import PPO
from garage.tf.baselines import ContinuousMLPBaselineWithModel
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.policies import GaussianLSTMPolicy, GaussianMLPPolicy
from tests.fixtures import snapshot_config, TfGraphTestCase


class TestPPO(TfGraphTestCase):

    @pytest.mark.huge
    def test_ppo_pendulum_continuous_baseline(self):
        """Test PPO with Pendulum environment."""
        with LocalTFRunner(snapshot_config, sess=self.sess) as runner:
            env = TfEnv(normalize(gym.make('InvertedDoublePendulum-v2')))
            policy = GaussianMLPPolicy(
                env_spec=env.spec,
                hidden_sizes=(64, 64),
                hidden_nonlinearity=tf.nn.tanh,
                output_nonlinearity=None,
            )
            baseline = ContinuousMLPBaselineWithModel(
                env_spec=env.spec,
                regressor_args=dict(hidden_sizes=(32, 32)),
            )
            algo = PPO(
                env_spec=env.spec,
                policy=policy,
                baseline=baseline,
                max_path_length=100,
                discount=0.99,
                gae_lambda=0.95,
                lr_clip_range=0.2,
                optimizer_args=dict(
                    batch_size=32,
                    max_epochs=10,
                ),
                stop_entropy_gradient=True,
                entropy_method='max',
                policy_ent_coeff=0.02,
                center_adv=False,
            )
            runner.setup(algo, env)
            last_avg_ret = runner.train(n_epochs=10, batch_size=2048)
            assert last_avg_ret > 100

            env.close()

    @pytest.mark.large
    def test_ppo_pendulum_recurrent_continuous_baseline(self):
        """Test PPO with Pendulum environment and recurrent policy."""
        with LocalTFRunner(snapshot_config) as runner:
            env = TfEnv(normalize(gym.make('InvertedDoublePendulum-v2')))
            policy = GaussianLSTMPolicy(env_spec=env.spec, )
            baseline = ContinuousMLPBaselineWithModel(
                env_spec=env.spec,
                regressor_args=dict(hidden_sizes=(32, 32)),
            )
            algo = PPO(
                env_spec=env.spec,
                policy=policy,
                baseline=baseline,
                max_path_length=100,
                discount=0.99,
                gae_lambda=0.95,
                lr_clip_range=0.2,
                optimizer_args=dict(
                    batch_size=32,
                    max_epochs=10,
                ),
                stop_entropy_gradient=True,
                entropy_method='max',
                policy_ent_coeff=0.02,
                center_adv=False,
            )
            runner.setup(algo, env)
            last_avg_ret = runner.train(n_epochs=10, batch_size=2048)
            assert last_avg_ret > 100

            env.close()



from unittest import mock

import numpy as np
import pytest
import torch

from garage.torch.policies import GaussianMLPPolicy

test_settings = [
    (1, 1, (1, )),
    (1, 2, (2, )),
    (1, 3, (3, )),
    (1, 1, (1, 2)),
    (1, 2, (2, 1)),
    (1, 3, (4, 5)),
    (2, 1, (1, )),
    (2, 2, (2, )),
    (2, 3, (3, )),
    (2, 1, (1, 2)),
    (2, 2, (2, 1)),
    (2, 3, (4, 5)),
    (5, 1, (1, )),
    (5, 2, (2, )),
    (5, 3, (3, )),
    (5, 1, (1, 2)),
    (5, 2, (2, 1)),
    (5, 3, (4, 5)),
]


@pytest.mark.parametrize('input_dim, output_dim, hidden_sizes', test_settings)
@mock.patch('garage.torch.modules.GaussianMLPModule')
def test_policy_get_actions(mock_model, input_dim, output_dim, hidden_sizes):
    action = torch.randn((output_dim, ))

    mock_dist = mock.MagicMock()
    mock_dist.rsample.return_value = action

    mock_model.return_value = mock_dist

    env_spec = mock.MagicMock()
    env_spec.observation_space.flat_dim = input_dim
    env_spec.action_space.flat_dim = output_dim

    policy = GaussianMLPPolicy(env_spec, mock_model)

    input = torch.ones(input_dim)
    sample = policy.get_actions(input)

    assert np.array_equal(sample, action.detach().numpy())


@pytest.mark.parametrize('input_dim, output_dim, hidden_sizes', test_settings)
@mock.patch('garage.torch.modules.GaussianMLPModule')
def test_policy_get_action(mock_model, input_dim, output_dim, hidden_sizes):
    action = torch.randn((
        1,
        output_dim,
    ))

    mock_dist = mock.MagicMock()
    mock_dist.rsample.return_value = action

    mock_model.return_value = mock_dist

    env_spec = mock.MagicMock()
    env_spec.observation_space.flat_dim = input_dim
    env_spec.action_space.flat_dim = output_dim

    policy = GaussianMLPPolicy(env_spec, mock_model)

    input = torch.ones(input_dim)
    sample = policy.get_action(input)

    assert np.array_equal(sample, np.squeeze(action.detach().numpy()))



from dowel import tabular
import numpy as np
import tensorflow as tf

from garage.core import Serializable
from garage.tf.core import LayersPowered, Parameterized
import garage.tf.core.layers as L
from garage.tf.core.network import MLP
from garage.tf.distributions import Categorical
from garage.tf.misc import tensor_utils
from garage.tf.optimizers import ConjugateGradientOptimizer, LbfgsOptimizer

NONE = list()


class CategoricalMLPRegressor(LayersPowered, Serializable, Parameterized):
    """
    A class for performing regression (or classification, really) by fitting a
    categorical distribution to the outputs. Assumes that the outputs will be
    always a one hot vector.
    """

    def __init__(
            self,
            input_shape,
            output_dim,
            name='CategoricalMLPRegressor',
            prob_network=None,
            hidden_sizes=(32, 32),
            hidden_nonlinearity=tf.nn.tanh,
            optimizer=None,
            tr_optimizer=None,
            use_trust_region=True,
            max_kl_step=0.01,
            normalize_inputs=True,
            no_initial_trust_region=True,
    ):
        """
        :param input_shape: Shape of the input data.
        :param output_dim: Dimension of output.
        :param hidden_sizes: Number of hidden units of each layer of the mean
        network.
        :param hidden_nonlinearity: Non-linearity used for each layer of the
        mean network.
        :param optimizer: Optimizer for minimizing the negative log-likelihood.
        :param use_trust_region: Whether to use trust region constraint.
        :param max_kl_step: KL divergence constraint for each iteration
        """
        Parameterized.__init__(self)
        Serializable.quick_init(self, locals())

        with tf.compat.v1.variable_scope(name, 'CategoricalMLPRegressor'):
            if optimizer is None:
                optimizer = LbfgsOptimizer()
            if tr_optimizer is None:
                tr_optimizer = ConjugateGradientOptimizer()

            self.output_dim = output_dim
            self.optimizer = optimizer
            self.tr_optimizer = tr_optimizer

            self._prob_network_name = 'prob_network'
            if prob_network is None:
                prob_network = MLP(
                    input_shape=input_shape,
                    output_dim=output_dim,
                    hidden_sizes=hidden_sizes,
                    hidden_nonlinearity=hidden_nonlinearity,
                    output_nonlinearity=tf.nn.softmax,
                    name=self._prob_network_name)

            l_prob = prob_network.output_layer

            LayersPowered.__init__(self, [l_prob])

            xs_var = prob_network.input_layer.input_var
            ys_var = tf.compat.v1.placeholder(
                dtype=tf.float32, shape=[None, output_dim], name='ys')
            old_prob_var = tf.compat.v1.placeholder(
                dtype=tf.float32, shape=[None, output_dim], name='old_prob')

            x_mean_var = tf.compat.v1.get_variable(
                name='x_mean',
                shape=(1, ) + input_shape,
                initializer=tf.constant_initializer(0., dtype=tf.float32))
            x_std_var = tf.compat.v1.get_variable(
                name='x_std',
                shape=(1, ) + input_shape,
                initializer=tf.constant_initializer(1., dtype=tf.float32))

            normalized_xs_var = (xs_var - x_mean_var) / x_std_var

            with tf.name_scope(
                    self._prob_network_name, values=[normalized_xs_var]):
                prob_var = L.get_output(
                    l_prob, {prob_network.input_layer: normalized_xs_var})

            old_info_vars = dict(prob=old_prob_var)
            info_vars = dict(prob=prob_var)

            dist = self._dist = Categorical(output_dim)

            mean_kl = tf.reduce_mean(dist.kl_sym(old_info_vars, info_vars))

            loss = -tf.reduce_mean(dist.log_likelihood_sym(ys_var, info_vars))

            predicted = tf.one_hot(
                tf.argmax(prob_var, axis=1), depth=output_dim)

            self.prob_network = prob_network
            self.f_predict = tensor_utils.compile_function([xs_var], predicted)
            self.f_prob = tensor_utils.compile_function([xs_var], prob_var)
            self.l_prob = l_prob

            self.optimizer.update_opt(
                loss=loss,
                target=self,
                network_outputs=[prob_var],
                inputs=[xs_var, ys_var])
            self.tr_optimizer.update_opt(
                loss=loss,
                target=self,
                network_outputs=[prob_var],
                inputs=[xs_var, ys_var, old_prob_var],
                leq_constraint=(mean_kl, max_kl_step))

            self.use_trust_region = use_trust_region
            self.name = name

            self.normalize_inputs = normalize_inputs
            self.x_mean_var = x_mean_var
            self.x_std_var = x_std_var
            self.first_optimized = not no_initial_trust_region

    def fit(self, xs, ys):
        if self.normalize_inputs:
            # recompute normalizing constants for inputs
            new_mean = np.mean(xs, axis=0, keepdims=True)
            new_std = np.std(xs, axis=0, keepdims=True) + 1e-8
            tf.compat.v1.get_default_session().run(
                tf.group(
                    tf.compat.v1.assign(self.x_mean_var, new_mean),
                    tf.compat.v1.assign(self.x_std_var, new_std),
                ))
        if self.use_trust_region and self.first_optimized:
            old_prob = self.f_prob(xs)
            inputs = [xs, ys, old_prob]
            optimizer = self.tr_optimizer
        else:
            inputs = [xs, ys]
            optimizer = self.optimizer
        loss_before = optimizer.loss(inputs)
        if self.name:
            prefix = self.name + '/'
        else:
            prefix = ''
        tabular.record(prefix + 'LossBefore', loss_before)
        optimizer.optimize(inputs)
        loss_after = optimizer.loss(inputs)
        tabular.record(prefix + 'LossAfter', loss_after)
        tabular.record(prefix + 'dLoss', loss_before - loss_after)
        self.first_optimized = True

    def predict(self, xs):
        return self.f_predict(np.asarray(xs))

    def predict_log_likelihood(self, xs, ys):
        prob = self.f_prob(np.asarray(xs))
        return self._dist.log_likelihood(np.asarray(ys), dict(prob=prob))

    def dist_info_sym(self, x_var, name=None):
        with tf.name_scope(name, 'dist_info_sym', [x_var]):
            normalized_xs_var = (x_var - self.x_mean_var) / self.x_std_var
            with tf.name_scope(
                    self._prob_network_name, values=[normalized_xs_var]):
                prob = L.get_output(
                    self.l_prob,
                    {self.prob_network.input_layer: normalized_xs_var})
        return dict(prob=prob)

    def log_likelihood_sym(self, x_var, y_var, name=None):
        with tf.name_scope(name, 'log_likelihood_sym', [x_var, y_var]):
            normalized_xs_var = (x_var - self.x_mean_var) / self.x_std_var
            with tf.name_scope(
                    self._prob_network_name, values=[normalized_xs_var]):
                prob = L.get_output(
                    self.l_prob,
                    {self.prob_network.input_layer: normalized_xs_var})
        return self._dist.log_likelihood_sym(y_var, dict(prob=prob))



"""Policies for TensorFlow-based algorithms."""
from garage.tf.policies.base import Policy
from garage.tf.policies.base import StochasticPolicy
from garage.tf.policies.categorical_conv_policy import CategoricalConvPolicy
from garage.tf.policies.categorical_conv_policy_with_model import (
    CategoricalConvPolicyWithModel)
from garage.tf.policies.categorical_gru_policy import CategoricalGRUPolicy
from garage.tf.policies.categorical_gru_policy_with_model import (
    CategoricalGRUPolicyWithModel)
from garage.tf.policies.categorical_lstm_policy import CategoricalLSTMPolicy
from garage.tf.policies.categorical_lstm_policy_with_model import (
    CategoricalLSTMPolicyWithModel)
from garage.tf.policies.categorical_mlp_policy import CategoricalMLPPolicy
from garage.tf.policies.categorical_mlp_policy_with_model import (
    CategoricalMLPPolicyWithModel)
from garage.tf.policies.continuous_mlp_policy import ContinuousMLPPolicy
from garage.tf.policies.continuous_mlp_policy_with_model import (
    ContinuousMLPPolicyWithModel)
from garage.tf.policies.deterministic_mlp_policy import DeterministicMLPPolicy
from garage.tf.policies.discrete_qf_derived_policy import (
    DiscreteQfDerivedPolicy)
from garage.tf.policies.gaussian_gru_policy import GaussianGRUPolicy
from garage.tf.policies.gaussian_gru_policy_with_model import (
    GaussianGRUPolicyWithModel)
from garage.tf.policies.gaussian_lstm_policy import GaussianLSTMPolicy
from garage.tf.policies.gaussian_lstm_policy_with_model import (
    GaussianLSTMPolicyWithModel)
from garage.tf.policies.gaussian_mlp_policy import GaussianMLPPolicy

__all__ = [
    'Policy', 'StochasticPolicy', 'CategoricalConvPolicy',
    'CategoricalConvPolicyWithModel', 'CategoricalGRUPolicy',
    'CategoricalGRUPolicyWithModel', 'CategoricalLSTMPolicy',
    'CategoricalLSTMPolicyWithModel', 'CategoricalMLPPolicy',
    'CategoricalMLPPolicyWithModel', 'ContinuousMLPPolicy',
    'ContinuousMLPPolicyWithModel', 'DiscreteQfDerivedPolicy',
    'DeterministicMLPPolicy', 'GaussianGRUPolicy', 'GaussianLSTMPolicy',
    'GaussianGRUPolicyWithModel', 'GaussianLSTMPolicyWithModel',
    'GaussianMLPPolicy'
]



"""
This is an integration test to make sure scripts from examples/
work when running `python examples/xx/xxx.py`.
"""
from garage.experiment import LocalRunner, run_experiment
from garage.np.baselines import LinearFeatureBaseline
from garage.tf.algos import VPG
from garage.tf.envs import TfEnv
from garage.tf.policies import CategoricalMLPPolicy


def _run_task(snapshot_config, *_):
    with LocalRunner(snapshot_config=snapshot_config) as runner:
        env = TfEnv(env_name='CartPole-v1')

        policy = CategoricalMLPPolicy(name='policy',
                                      env_spec=env.spec,
                                      hidden_sizes=(32, 32))

        baseline = LinearFeatureBaseline(env_spec=env.spec)

        algo = VPG(
            env_spec=env.spec,
            policy=policy,
            baseline=baseline,
            max_path_length=100,
            discount=0.99,
            optimizer_args=dict(tf_optimizer_args=dict(learning_rate=0.01, )))

        runner.setup(algo, env)
        runner.train(n_epochs=3, batch_size=100)


if __name__ == '__main__':
    run_experiment(
        _run_task,
        snapshot_mode='last',
        seed=1,
    )



"""Base class for batch sampling-based policy optimization methods."""

import collections

from dowel import logger, tabular
import numpy as np

from garage.misc import special
from garage.np.algos import RLAlgorithm
from garage.sampler import OnPolicyVectorizedSampler
from garage.tf.misc import tensor_utils
from garage.tf.samplers import BatchSampler


class BatchPolopt(RLAlgorithm):
    """Base class for batch sampling-based policy optimization methods.

    This includes various policy gradient methods like VPG, NPG, PPO, TRPO,
    etc.

    Args:
        env_spec (garage.envs.EnvSpec): Environment specification.
        policy (garage.tf.policies.base.Policy): Policy.
        baseline (garage.tf.baselines.Baseline): The baseline.
        scope (str): Scope for identifying the algorithm.
            Must be specified if running multiple algorithms
            simultaneously, each using different environments
            and policies.
        max_path_length (int): Maximum length of a single rollout.
        discount (float): Discount.
        gae_lambda (float): Lambda used for generalized advantage
            estimation.
        center_adv (bool): Whether to rescale the advantages
            so that they have mean 0 and standard deviation 1.
        positive_adv (bool): Whether to shift the advantages
            so that they are always positive. When used in
            conjunction with center_adv the advantages will be
            standardized before shifting.
        fixed_horizon (bool): Whether to fix horizon.

    """

    def __init__(self,
                 env_spec,
                 policy,
                 baseline,
                 scope=None,
                 max_path_length=500,
                 discount=0.99,
                 gae_lambda=1,
                 center_adv=True,
                 positive_adv=False,
                 fixed_horizon=False):
        self.env_spec = env_spec
        self.policy = policy
        self.baseline = baseline
        self.scope = scope
        self.max_path_length = max_path_length
        self.discount = discount
        self.gae_lambda = gae_lambda
        self.center_adv = center_adv
        self.positive_adv = positive_adv
        self.fixed_horizon = fixed_horizon

        self.episode_reward_mean = collections.deque(maxlen=100)
        if policy.vectorized:
            self.sampler_cls = OnPolicyVectorizedSampler
        else:
            self.sampler_cls = BatchSampler

        self.init_opt()

    def train(self, runner):
        """Obtain samplers and start actual training for each epoch.

        Args:
            runner (LocalRunner): LocalRunner is passed to give algorithm
                the access to runner.step_epochs(), which provides services
                such as snapshotting and sampler control.

        Returns:
            The average return in last epoch cycle.

        """
        last_return = None

        for epoch in runner.step_epochs():
            runner.step_path = runner.obtain_samples(runner.step_itr)
            last_return = self.train_once(runner.step_itr, runner.step_path)
            runner.step_itr += 1

        return last_return

    def train_once(self, itr, paths):
        """Perform one step of policy optimization given one batch of samples.

        Args:
            itr (int): Iteration number.
            paths (list[dict]): A list of collected paths.

        """
        paths = self.process_samples(itr, paths)

        self.log_diagnostics(paths)
        logger.log('Optimizing policy...')
        self.optimize_policy(itr, paths)
        return paths['average_return']

    def log_diagnostics(self, paths):
        """Log diagnostic information."""
        logger.log('Logging diagnostics...')
        self.policy.log_diagnostics(paths)
        self.baseline.log_diagnostics(paths)

    def process_samples(self, itr, paths):
        """Return processed sample data based on the collected paths.

        Args:
            itr (int): Iteration number.
            paths (list[dict]): A list of collected paths.

        Returns:
            dict: Processed sample data, with key
                * observations: (numpy.ndarray)
                * actions: (numpy.ndarray)
                * rewards: (numpy.ndarray)
                * baselines: (numpy.ndarray)
                * returns: (numpy.ndarray)
                * valids: (numpy.ndarray)
                * agent_infos: (dict)
                * env_infos: (dict)
                * paths: (list[dict])
                * average_return: (numpy.float64)

        """
        baselines = []
        returns = []

        max_path_length = self.max_path_length

        if hasattr(self.baseline, 'predict_n'):
            all_path_baselines = self.baseline.predict_n(paths)
        else:
            all_path_baselines = [
                self.baseline.predict(path) for path in paths
            ]

        for idx, path in enumerate(paths):
            path_baselines = np.append(all_path_baselines[idx], 0)
            deltas = (path['rewards'] + self.discount * path_baselines[1:] -
                      path_baselines[:-1])
            path['advantages'] = special.discount_cumsum(
                deltas, self.discount * self.gae_lambda)
            path['deltas'] = deltas

        for idx, path in enumerate(paths):
            # baselines
            path['baselines'] = all_path_baselines[idx]
            baselines.append(path['baselines'])

            # returns
            path['returns'] = special.discount_cumsum(path['rewards'],
                                                      self.discount)
            returns.append(path['returns'])

        # make all paths the same length
        obs = [path['observations'] for path in paths]
        obs = tensor_utils.pad_tensor_n(obs, max_path_length)

        actions = [path['actions'] for path in paths]
        actions = tensor_utils.pad_tensor_n(actions, max_path_length)

        rewards = [path['rewards'] for path in paths]
        rewards = tensor_utils.pad_tensor_n(rewards, max_path_length)

        returns = [path['returns'] for path in paths]
        returns = tensor_utils.pad_tensor_n(returns, max_path_length)

        baselines = tensor_utils.pad_tensor_n(baselines, max_path_length)

        agent_infos = [path['agent_infos'] for path in paths]
        agent_infos = tensor_utils.stack_tensor_dict_list([
            tensor_utils.pad_tensor_dict(p, max_path_length)
            for p in agent_infos
        ])

        env_infos = [path['env_infos'] for path in paths]
        env_infos = tensor_utils.stack_tensor_dict_list([
            tensor_utils.pad_tensor_dict(p, max_path_length) for p in env_infos
        ])

        valids = [np.ones_like(path['returns']) for path in paths]
        valids = tensor_utils.pad_tensor_n(valids, max_path_length)

        average_discounted_return = (np.mean(
            [path['returns'][0] for path in paths]))

        undiscounted_returns = [sum(path['rewards']) for path in paths]
        self.episode_reward_mean.extend(undiscounted_returns)

        ent = np.sum(self.policy.distribution.entropy(agent_infos) *
                     valids) / np.sum(valids)

        samples_data = dict(
            observations=obs,
            actions=actions,
            rewards=rewards,
            baselines=baselines,
            returns=returns,
            valids=valids,
            agent_infos=agent_infos,
            env_infos=env_infos,
            paths=paths,
            average_return=np.mean(undiscounted_returns),
        )

        tabular.record('Iteration', itr)
        tabular.record('AverageDiscountedReturn', average_discounted_return)
        tabular.record('AverageReturn', np.mean(undiscounted_returns))
        tabular.record('Extras/EpisodeRewardMean',
                       np.mean(self.episode_reward_mean))
        tabular.record('NumTrajs', len(paths))
        tabular.record('Entropy', ent)
        tabular.record('Perplexity', np.exp(ent))
        tabular.record('StdReturn', np.std(undiscounted_returns))
        tabular.record('MaxReturn', np.max(undiscounted_returns))
        tabular.record('MinReturn', np.min(undiscounted_returns))

        return samples_data

    def init_opt(self):
        """Initialize the optimization procedure.

        If using tensorflow, this may include declaring all the variables and
        compiling functions.
        """
        raise NotImplementedError

    def get_itr_snapshot(self, itr):
        """Get all the data that should be saved in this snapshot iteration."""
        raise NotImplementedError

    def optimize_policy(self, itr, samples_data):
        """Optimize the policy using the samples."""
        raise NotImplementedError



"""This module implements continuous mlp baseline."""
import numpy as np

from garage.core import Serializable
from garage.misc.overrides import overrides
from garage.np.baselines import Baseline
from garage.tf.core import Parameterized
from garage.tf.regressors import ContinuousMLPRegressor


class ContinuousMLPBaseline(Baseline, Parameterized, Serializable):
    """A value function using a mlp network."""

    def __init__(
            self,
            env_spec,
            subsample_factor=1.,
            num_seq_inputs=1,
            regressor_args=None,
            name='ContinuousMLPBaseline',
    ):
        """
        Constructor.

        :param env_spec: environment specification.
        :param subsample_factor:
        :param num_seq_inputs: number of sequence inputs.
        :param regressor_args: regressor arguments.
        """
        Parameterized.__init__(self)
        Serializable.quick_init(self, locals())
        super(ContinuousMLPBaseline, self).__init__(env_spec)
        if regressor_args is None:
            regressor_args = dict()

        self._regressor = ContinuousMLPRegressor(
            input_shape=(
                env_spec.observation_space.flat_dim * num_seq_inputs, ),
            output_dim=1,
            name=name,
            **regressor_args)
        self.name = name

    @overrides
    def get_param_values(self, **tags):
        """Get parameter values."""
        return self._regressor.get_param_values(**tags)

    @overrides
    def set_param_values(self, val, **tags):
        """Set parameter values to val."""
        self._regressor.set_param_values(val, **tags)

    @overrides
    def fit(self, paths):
        """Fit regressor based on paths."""
        observations = np.concatenate([p['observations'] for p in paths])
        returns = np.concatenate([p['returns'] for p in paths])
        self._regressor.fit(observations, returns.reshape((-1, 1)))

    @overrides
    def predict(self, path):
        """Predict value based on paths."""
        return self._regressor.predict(path['observations']).flatten()

    @overrides
    def get_params_internal(self, **tags):
        return self._regressor.get_params_internal(**tags)



import gym
import pytest

from garage.tf.envs.base import TfEnv
from tests.helpers import step_env_with_gym_quirks


class TestTfEnv:
    @pytest.mark.nightly
    @pytest.mark.parametrize('spec', list(gym.envs.registry.all()))
    def test_all_gym_envs(self, spec):
        env = TfEnv(spec.make())
        step_env_with_gym_quirks(env, spec)

    @pytest.mark.nightly
    @pytest.mark.parametrize('spec', list(gym.envs.registry.all()))
    def test_all_gym_envs_pickleable(self, spec):
        env = TfEnv(env_name=spec.id)
        step_env_with_gym_quirks(
            env, spec, n=1, render=True, serialize_env=True)



import pickle
from unittest import mock

import numpy as np
import pytest
import tensorflow as tf

from garage.tf.models import GaussianLSTMModel
from tests.fixtures import TfGraphTestCase
from tests.helpers import recurrent_step_lstm


class TestGaussianLSTMModel(TfGraphTestCase):
    def setup_method(self):
        super().setup_method()
        self.batch_size = 1
        self.time_step = 2
        self.feature_shape = 2
        self.default_initializer = tf.constant_initializer(0.1)

        self.obs_inputs = np.full(
            (self.batch_size, self.time_step, self.feature_shape), 1.)
        self.obs_input = np.full((self.batch_size, self.feature_shape), 1.)

        self.input_var = tf.compat.v1.placeholder(
            tf.float32, shape=(None, None, self.feature_shape), name='input')
        self.step_input_var = tf.compat.v1.placeholder(
            tf.float32, shape=(None, self.feature_shape), name='step_input')

    # yapf: disable
    @pytest.mark.parametrize('output_dim, hidden_dim', [
        (1, 1),
        (2, 2),
        (3, 3)
    ])
    # yapf: enable
    @mock.patch('tensorflow.random.normal')
    def test_std_share_network_output_values(self, mock_normal, output_dim,
                                             hidden_dim):
        mock_normal.return_value = 0.5
        model = GaussianLSTMModel(
            output_dim=output_dim,
            hidden_dim=hidden_dim,
            std_share_network=True,
            hidden_nonlinearity=None,
            recurrent_nonlinearity=None,
            hidden_w_init=self.default_initializer,
            recurrent_w_init=self.default_initializer,
            output_w_init=self.default_initializer)
        step_hidden_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, hidden_dim),
            name='step_hidden',
            dtype=tf.float32)
        step_cell_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, hidden_dim),
            name='step_cell',
            dtype=tf.float32)
        (action_var, mean_var, step_mean_var, log_std_var, step_log_std_var,
         step_hidden, step_cell, hidden_init_var, cell_init_var,
         dist) = model.build(self.input_var, self.step_input_var,
                             step_hidden_var, step_cell_var)

        hidden1 = hidden2 = np.full((self.batch_size, hidden_dim),
                                    hidden_init_var.eval())
        cell1 = cell2 = np.full((self.batch_size, hidden_dim),
                                cell_init_var.eval())

        mean, log_std = self.sess.run(
            [mean_var, log_std_var],
            feed_dict={self.input_var: self.obs_inputs})

        for i in range(self.time_step):
            action, mean1, log_std1, hidden1, cell1 = self.sess.run(
                [
                    action_var, step_mean_var, step_log_std_var, step_hidden,
                    step_cell
                ],
                feed_dict={
                    self.step_input_var: self.obs_input,
                    step_hidden_var: hidden1,
                    step_cell_var: cell1
                })

            hidden2, cell2 = recurrent_step_lstm(
                input_val=self.obs_input,
                num_units=hidden_dim,
                step_hidden=hidden2,
                step_cell=cell2,
                w_x_init=0.1,
                w_h_init=0.1,
                b_init=0.,
                nonlinearity=None,
                gate_nonlinearity=None)

            output_nonlinearity = np.full(
                (np.prod(hidden2.shape[1:]), output_dim), 0.1)
            output2 = np.matmul(hidden2, output_nonlinearity)
            assert np.allclose(mean1, output2)
            assert np.allclose(log_std1, output2)
            assert np.allclose(hidden1, hidden2)
            assert np.allclose(cell1, cell2)

            expected_action = 0.5 * np.exp(log_std1) + mean1
            assert np.allclose(action, expected_action)

    # yapf: disable
    @pytest.mark.parametrize('output_dim, hidden_dim', [
        (1, 1),
        (2, 2),
        (3, 3)
    ])
    # yapf: enable
    def test_std_share_network_shapes(self, output_dim, hidden_dim):
        model = GaussianLSTMModel(
            output_dim=output_dim,
            hidden_dim=hidden_dim,
            std_share_network=True,
            hidden_nonlinearity=None,
            recurrent_nonlinearity=None,
            hidden_w_init=self.default_initializer,
            recurrent_w_init=self.default_initializer,
            output_w_init=self.default_initializer)
        step_hidden_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, hidden_dim),
            name='step_hidden',
            dtype=tf.float32)
        step_cell_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, hidden_dim),
            name='step_cell',
            dtype=tf.float32)
        (action_var, mean_var, step_mean_var, log_std_var, step_log_std_var,
         step_hidden, step_cell, hidden_init_var, cell_init_var,
         dist) = model.build(self.input_var, self.step_input_var,
                             step_hidden_var, step_cell_var)

        # output layer is a tf.keras.layers.Dense object,
        # which cannot be access by tf.compat.v1.variable_scope.
        # A workaround is to access in tf.compat.v1.global_variables()
        for var in tf.compat.v1.global_variables():
            if 'output_layer/kernel' in var.name:
                std_share_output_weights = var
            if 'output_layer/bias' in var.name:
                std_share_output_bias = var
        assert std_share_output_weights.shape[1] == output_dim * 2
        assert std_share_output_bias.shape == output_dim * 2

    @pytest.mark.parametrize('output_dim, hidden_dim, init_std', [(1, 1, 1),
                                                                  (1, 1, 2),
                                                                  (1, 2, 1),
                                                                  (1, 2, 2),
                                                                  (3, 3, 1),
                                                                  (3, 3, 2)])
    @mock.patch('tensorflow.random.normal')
    def test_without_std_share_network_output_values(
            self, mock_normal, output_dim, hidden_dim, init_std):
        mock_normal.return_value = 0.5
        model = GaussianLSTMModel(
            output_dim=output_dim,
            hidden_dim=hidden_dim,
            std_share_network=False,
            hidden_nonlinearity=None,
            recurrent_nonlinearity=None,
            hidden_w_init=self.default_initializer,
            recurrent_w_init=self.default_initializer,
            output_w_init=self.default_initializer,
            init_std=init_std)
        step_hidden_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, hidden_dim),
            name='step_hidden',
            dtype=tf.float32)
        step_cell_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, hidden_dim),
            name='step_cell',
            dtype=tf.float32)
        (action_var, mean_var, step_mean_var, log_std_var, step_log_std_var,
         step_hidden, step_cell, hidden_init_var, cell_init_var,
         dist) = model.build(self.input_var, self.step_input_var,
                             step_hidden_var, step_cell_var)

        hidden1 = hidden2 = np.full((self.batch_size, hidden_dim),
                                    hidden_init_var.eval())
        cell1 = cell2 = np.full((self.batch_size, hidden_dim),
                                cell_init_var.eval())

        mean, log_std = self.sess.run(
            [mean_var, log_std_var],
            feed_dict={self.input_var: self.obs_inputs})

        for i in range(self.time_step):
            action, mean1, log_std1, hidden1, cell1 = self.sess.run(
                [
                    action_var, step_mean_var, step_log_std_var, step_hidden,
                    step_cell
                ],
                feed_dict={
                    self.step_input_var: self.obs_input,
                    step_hidden_var: hidden1,
                    step_cell_var: cell1
                })

            hidden2, cell2 = recurrent_step_lstm(
                input_val=self.obs_input,
                num_units=hidden_dim,
                step_hidden=hidden2,
                step_cell=cell2,
                w_x_init=0.1,
                w_h_init=0.1,
                b_init=0.,
                nonlinearity=None,
                gate_nonlinearity=None)

            output_nonlinearity = np.full(
                (np.prod(hidden2.shape[1:]), output_dim), 0.1)
            output2 = np.matmul(hidden2, output_nonlinearity)
            assert np.allclose(mean1, output2)
            expected_log_std = np.full((self.batch_size, output_dim),
                                       np.log(init_std))
            assert np.allclose(log_std1, expected_log_std)
            assert np.allclose(hidden1, hidden2)
            assert np.allclose(cell1, cell2)

            expected_action = 0.5 * np.exp(log_std1) + mean1
            assert np.allclose(action, expected_action)

    # yapf: disable
    @pytest.mark.parametrize('output_dim, hidden_dim', [
        (1, 1),
        (2, 2),
        (3, 3)
    ])
    # yapf: enable
    def test_without_std_share_network_shapes(self, output_dim, hidden_dim):
        model = GaussianLSTMModel(
            output_dim=output_dim,
            hidden_dim=hidden_dim,
            std_share_network=False,
            hidden_nonlinearity=None,
            recurrent_nonlinearity=None,
            hidden_w_init=self.default_initializer,
            recurrent_w_init=self.default_initializer,
            output_w_init=self.default_initializer)
        step_hidden_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, hidden_dim),
            name='step_hidden',
            dtype=tf.float32)
        step_cell_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, hidden_dim),
            name='step_cell',
            dtype=tf.float32)
        (action_var, mean_var, step_mean_var, log_std_var, step_log_std_var,
         step_hidden, step_cell, hidden_init_var, cell_init_var,
         dist) = model.build(self.input_var, self.step_input_var,
                             step_hidden_var, step_cell_var)

        # output layer is a tf.keras.layers.Dense object,
        # which cannot be access by tf.compat.v1.variable_scope.
        # A workaround is to access in tf.compat.v1.global_variables()
        for var in tf.compat.v1.global_variables():
            if 'output_layer/kernel' in var.name:
                std_share_output_weights = var
            if 'output_layer/bias' in var.name:
                std_share_output_bias = var
            if 'log_std_param/parameter' in var.name:
                log_std_param = var
        assert std_share_output_weights.shape[1] == output_dim
        assert std_share_output_bias.shape == output_dim
        assert log_std_param.shape == output_dim

    # yapf: disable
    @pytest.mark.parametrize('output_dim, hidden_dim', [
        (1, 1),
        (2, 2),
        (3, 3)
    ])
    # yapf: enable
    @mock.patch('tensorflow.random.normal')
    def test_std_share_network_is_pickleable(self, mock_normal, output_dim,
                                             hidden_dim):
        mock_normal.return_value = 0.5
        model = GaussianLSTMModel(
            output_dim=output_dim,
            hidden_dim=hidden_dim,
            std_share_network=True,
            hidden_nonlinearity=None,
            recurrent_nonlinearity=None,
            hidden_w_init=self.default_initializer,
            recurrent_w_init=self.default_initializer,
            output_w_init=self.default_initializer)
        step_hidden_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, hidden_dim),
            name='step_hidden',
            dtype=tf.float32)
        step_cell_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, hidden_dim),
            name='step_cell',
            dtype=tf.float32)
        (_, mean_var, step_mean_var, log_std_var, step_log_std_var,
         step_hidden, step_cell, _, _, _) = model.build(
             self.input_var, self.step_input_var, step_hidden_var,
             step_cell_var)

        # output layer is a tf.keras.layers.Dense object,
        # which cannot be access by tf.compat.v1.variable_scope.
        # A workaround is to access in tf.compat.v1.global_variables()
        for var in tf.compat.v1.global_variables():
            if 'output_layer/bias' in var.name:
                var.load(tf.ones_like(var).eval())

        hidden = np.zeros((self.batch_size, hidden_dim))
        cell = np.zeros((self.batch_size, hidden_dim))

        outputs1 = self.sess.run([mean_var, log_std_var],
                                 feed_dict={self.input_var: self.obs_inputs})
        output1 = self.sess.run(
            [step_mean_var, step_log_std_var, step_hidden, step_cell],
            feed_dict={
                self.step_input_var: self.obs_input,
                step_hidden_var: hidden,
                step_cell_var: cell
            })

        h = pickle.dumps(model)
        with tf.compat.v1.Session(graph=tf.Graph()) as sess:
            model_pickled = pickle.loads(h)

            input_var = tf.compat.v1.placeholder(
                tf.float32,
                shape=(None, None, self.feature_shape),
                name='input')
            step_input_var = tf.compat.v1.placeholder(
                tf.float32,
                shape=(None, self.feature_shape),
                name='step_input')
            step_hidden_var = tf.compat.v1.placeholder(
                shape=(self.batch_size, hidden_dim),
                name='initial_hidden',
                dtype=tf.float32)
            step_cell_var = tf.compat.v1.placeholder(
                shape=(self.batch_size, hidden_dim),
                name='initial_cell',
                dtype=tf.float32)

            (_, mean_var2, step_mean_var2, log_std_var2, step_log_std_var2,
             step_hidden2, step_cell2, _, _, _) = model_pickled.build(
                 input_var, step_input_var, step_hidden_var, step_cell_var)

            outputs2 = sess.run([mean_var2, log_std_var2],
                                feed_dict={input_var: self.obs_inputs})
            output2 = sess.run(
                [step_mean_var2, step_log_std_var2, step_hidden2, step_cell2],
                feed_dict={
                    step_input_var: self.obs_input,
                    step_hidden_var: hidden,
                    step_cell_var: cell
                })
            assert np.array_equal(outputs1, outputs2)
            assert np.array_equal(output1, output2)

    # yapf: disable
    @pytest.mark.parametrize('output_dim, hidden_dim', [
        (1, 1),
        (2, 2),
        (3, 3)
    ])
    # yapf: enable
    @mock.patch('tensorflow.random.normal')
    def test_without_std_share_network_is_pickleable(self, mock_normal,
                                                     output_dim, hidden_dim):
        mock_normal.return_value = 0.5
        model = GaussianLSTMModel(
            output_dim=output_dim,
            hidden_dim=hidden_dim,
            std_share_network=False,
            hidden_nonlinearity=None,
            recurrent_nonlinearity=None,
            hidden_w_init=self.default_initializer,
            recurrent_w_init=self.default_initializer,
            output_w_init=self.default_initializer)
        step_hidden_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, hidden_dim),
            name='step_hidden',
            dtype=tf.float32)
        step_cell_var = tf.compat.v1.placeholder(
            shape=(self.batch_size, hidden_dim),
            name='step_cell',
            dtype=tf.float32)
        (_, mean_var, step_mean_var, log_std_var, step_log_std_var,
         step_hidden, step_cell, _, _, _) = model.build(
             self.input_var, self.step_input_var, step_hidden_var,
             step_cell_var)

        # output layer is a tf.keras.layers.Dense object,
        # which cannot be access by tf.compat.v1.variable_scope.
        # A workaround is to access in tf.compat.v1.global_variables()
        for var in tf.compat.v1.global_variables():
            if 'output_layer/bias' in var.name:
                var.load(tf.ones_like(var).eval())

        hidden = np.zeros((self.batch_size, hidden_dim))
        cell = np.zeros((self.batch_size, hidden_dim))

        outputs1 = self.sess.run([mean_var, log_std_var],
                                 feed_dict={self.input_var: self.obs_inputs})
        output1 = self.sess.run(
            [step_mean_var, step_log_std_var, step_hidden, step_cell],
            feed_dict={
                self.step_input_var: self.obs_input,
                step_hidden_var: hidden,
                step_cell_var: cell
            })

        h = pickle.dumps(model)
        with tf.compat.v1.Session(graph=tf.Graph()) as sess:
            model_pickled = pickle.loads(h)

            input_var = tf.compat.v1.placeholder(
                tf.float32,
                shape=(None, None, self.feature_shape),
                name='input')
            step_input_var = tf.compat.v1.placeholder(
                tf.float32,
                shape=(None, self.feature_shape),
                name='step_input')
            step_hidden_var = tf.compat.v1.placeholder(
                shape=(self.batch_size, hidden_dim),
                name='initial_hidden',
                dtype=tf.float32)
            step_cell_var = tf.compat.v1.placeholder(
                shape=(self.batch_size, hidden_dim),
                name='initial_cell',
                dtype=tf.float32)

            (_, mean_var2, step_mean_var2, log_std_var2, step_log_std_var2,
             step_hidden2, step_cell2, _, _, _) = model_pickled.build(
                 input_var, step_input_var, step_hidden_var, step_cell_var)

            outputs2 = sess.run([mean_var2, log_std_var2],
                                feed_dict={input_var: self.obs_inputs})
            output2 = sess.run(
                [step_mean_var2, step_log_std_var2, step_hidden2, step_cell2],
                feed_dict={
                    step_input_var: self.obs_input,
                    step_hidden_var: hidden,
                    step_cell_var: cell
                })
            assert np.array_equal(outputs1, outputs2)
            assert np.array_equal(output1, output2)



import pickle

import numpy as np
import pytest
import torch
import torch.nn as nn

from garage.torch.modules import MLPModule


class TestMLPModel:
    # yapf: disable
    @pytest.mark.parametrize('input_dim, output_dim, hidden_sizes', [
        (5, 1, (1, )),
        (5, 1, (2, )),
        (5, 2, (3, )),
        (5, 2, (1, 1)),
        (5, 3, (2, 2)),
    ])
    # yapf: enable
    def test_output_values(self, input_dim, output_dim, hidden_sizes):
        input_val = torch.ones([1, 5], dtype=torch.float32)
        module = MLPModule(
            input_dim=input_dim,
            output_dim=output_dim,
            hidden_nonlinearity=None,
            hidden_sizes=hidden_sizes,
            hidden_w_init=nn.init.ones_,
            output_w_init=nn.init.ones_)
        output = module(input_val)

        expected_output = torch.full([1, output_dim],
                                     fill_value=5 * np.prod(hidden_sizes),
                                     dtype=torch.float32)

        assert torch.all(torch.eq(output, expected_output))

    # yapf: disable
    @pytest.mark.parametrize('input_dim, output_dim, hidden_sizes', [
        (5, 1, (1, )),
        (5, 1, (2, )),
        (5, 2, (3, )),
        (5, 2, (1, 1)),
        (5, 3, (2, 2)),
    ])
    # yapf: enable
    def test_is_pickleable(self, input_dim, output_dim, hidden_sizes):
        input_val = torch.ones([1, 5], dtype=torch.float32)
        module = MLPModule(
            input_dim=input_dim,
            output_dim=output_dim,
            hidden_nonlinearity=None,
            hidden_sizes=hidden_sizes,
            hidden_w_init=nn.init.ones_,
            output_w_init=nn.init.ones_)
        output1 = module(input_val)

        h = pickle.dumps(module)
        model_pickled = pickle.loads(h)
        output2 = model_pickled(input_val)

        assert np.array_equal(torch.all(torch.eq(output1, output2)), True)



import pytest
import torch
from torch import nn

from garage.torch.modules import MultiHeadedMLPModule

plain_settings = [
    (1, 1, (1, ), (0, 1, 1, 2, 3, 5, 5)),
    (1, 2, (2, ), (0, 1, 2)),
    (1, 3, (3, ), (0, 1, 2, 3, 5, 5)),
    (1, 1, (1, 2), (0, 1, 1, 2, 3, 5, 5)),
    (1, 2, (2, 1), (0, 1, 1, 2, 5, 5)),
    (1, 3, (4, 5), (0, 1, 2, 3, 5, 5)),
    (2, 1, (1, ), (0, 1, 1, 3, 5, 5)),
    (2, 2, (2, ), (0, 1, 1, 2, 3, 5, 5)),
    (2, 3, (3, ), (1, 1, 2, 5, 5)),
    (2, 1, (1, 2), (0, 1, 2, 3, 5, 5)),
    (2, 2, (2, 1), (0, 1, 1, 2, 3, 5, 5)),
    (2, 3, (4, 5), (0, 1, 1, 2, 3, 5)),
    (5, 1, (1, ), (0, 1, 1, 2, 3, 5)),
    (5, 2, (2, ), (0, 1, 3, 5, 5)),
    (5, 3, (3, ), (0, 1, 1, 2, 3)),
    (5, 1, (1, 2), (0, 1, 1, 2, 5, 5)),
    (5, 2, (2, 1), (0, 1, 1, 2, 3, 5, 5)),
    (5, 3, (4, 5), (0, 1, 2, 3, 5, 5)),
]

different_output_dims_settings = [
    (1, (1, 4), (1, ), (0, 1)),
    (1, (2, 3), (2, ), (0, 3)),
    (1, (3, 2), (3, ), (5, )),
    (1, (1, 6), (1, 2), (6, 4)),
    (1, (2, 7), (2, 1), (5, 5)),
    (1, (3, 4), (4, 5), (1, 2)),
    (2, (1, 5, 6), (1, ), (6, 1, 3)),
    (2, (2, 1, 3), (2, ), (1, 2, 3)),
    (2, (3, 6, 7), (3, ), (6, )),
    (2, (1, ), (1, 2), (9, )),
    (2, (2, ), (2, 1), (4, )),
    (2, (3, ), (4, 5), (5, )),
    (5, (1, 3, 1), (1, ), (5, )),
    (5, (2, 6), (2, ), (4, 1)),
    (5, (3, 5), (3, ), (2, 5)),
    (5, (1, 3), (1, 2), (6, 8)),
    (5, (2, 1, 1, 1), (2, 1), (1, )),
    (5, (3, 4, 1, 2), (4, 5), (5, 1, 2, 3)),
]

invalid_settings = [
    (1, (1, 4, 5), (1, ), 2, (None, ), (1, 2),
     (nn.init.ones_, )),  # n_head != output_dims
    (1, (1, 4), (1, ), 2, (None, ), (1, 2, 3),
     (nn.init.ones_, )),  # n_head != w_init
    (1, (1, 4), (1, ), 2, (None, None, None), (1, 2),
     (nn.init.ones_, )),  # n_head != nonlinearity
    (1, (1, 4), (1, ), 2, (None, ), (1, 2),
     (nn.init.ones_, nn.init.ones_, nn.init.ones_)),  # n_head != b_init
    (1, (1, 4, 5), (1, ), 3, (None, ), (1, 2),
     (nn.init.ones_, )),  # output_dims > w_init
    (1, (1, ), (1, ), 1, (None, ), (1, 2, 3),
     (nn.init.ones_, )),  # output_dims < w_init
    (1, (1, 4, 5), (1, ), 3, (None, None), (1, 2, 3),
     (nn.init.ones_, )),  # output_dims > nonlinearity
    (1, (1, ), (1, ), 1, (None, None, None), (1, 2, 3),
     (nn.init.ones_, )),  # output_dims < nonlinearity
    (1, (1, 4, 5), (1, ), 3, (None, ), (1, 2, 3),
     (nn.init.ones_, nn.init.ones_)),  # output_dims > b_init
    (1, (1, ), (1, ), 1, (None, ), (1, 2, 3),
     (nn.init.ones_, nn.init.ones_, nn.init.ones_)),  # output_dims > b_init
]


def helper_make_inits(val):
    return lambda x: nn.init.constant_(x, val)


@pytest.mark.parametrize(
    'input_dim, output_dim, hidden_sizes, output_w_init_vals', plain_settings)
def test_multi_headed_mlp_module(input_dim, output_dim, hidden_sizes,
                                 output_w_init_vals):
    module = MultiHeadedMLPModule(
        n_heads=len(output_w_init_vals),
        input_dim=input_dim,
        output_dims=output_dim,
        hidden_sizes=hidden_sizes,
        hidden_nonlinearity=None,
        hidden_w_init=nn.init.ones_,
        output_nonlinearities=None,
        output_w_inits=list(map(helper_make_inits, output_w_init_vals)))

    input = torch.ones(input_dim)
    outputs = module(input)

    for i in range(len(outputs)):
        expected = input_dim * torch.Tensor(hidden_sizes).prod()
        expected *= output_w_init_vals[i]
        assert torch.equal(outputs[i], torch.full((output_dim, ), expected))


@pytest.mark.parametrize(
    'input_dim, output_dim, hidden_sizes, output_w_init_vals',
    different_output_dims_settings)
def test_multi_headed_mlp_module_with_different_output_dims(
        input_dim, output_dim, hidden_sizes, output_w_init_vals):
    module = MultiHeadedMLPModule(
        n_heads=len(output_dim),
        input_dim=input_dim,
        output_dims=output_dim,
        hidden_sizes=hidden_sizes,
        hidden_nonlinearity=None,
        hidden_w_init=nn.init.ones_,
        output_nonlinearities=None,
        output_w_inits=list(map(helper_make_inits, output_w_init_vals)))

    input = torch.ones(input_dim)
    outputs = module(input)

    if len(output_w_init_vals) == 1:
        output_w_init_vals = output_w_init_vals * len(output_dim)
    for i in range(len(outputs)):
        expected = input_dim * torch.Tensor(hidden_sizes).prod()
        expected *= output_w_init_vals[i]
        assert torch.equal(outputs[i], torch.full((output_dim[i], ), expected))


@pytest.mark.parametrize('input_dim, output_dim, hidden_sizes, '
                         'n_heads, nonlinearity, w_init, b_init',
                         invalid_settings)
def test_invalid_settings(input_dim, output_dim, hidden_sizes, n_heads,
                          nonlinearity, w_init, b_init):
    with pytest.raises(ValueError):
        MultiHeadedMLPModule(
            n_heads=n_heads,
            input_dim=input_dim,
            output_dims=output_dim,
            hidden_sizes=hidden_sizes,
            hidden_nonlinearity=None,
            hidden_w_init=nn.init.ones_,
            output_nonlinearities=nonlinearity,
            output_w_inits=list(map(helper_make_inits, w_init)),
            output_b_inits=b_init)



from contextlib import contextmanager

import tensorflow as tf

from garage.core import Serializable
from garage.misc.tensor_utils import flatten_tensors, unflatten_tensors

load_params = True


@contextmanager
def suppress_params_loading():
    global load_params
    load_params = False
    yield
    load_params = True


class Parameterized:
    def __init__(self):
        self._cached_params = {}
        self._cached_param_dtypes = {}
        self._cached_param_shapes = {}
        self._cached_assign_ops = {}
        self._cached_assign_placeholders = {}

    def get_params_internal(self, **tags):
        """
        Internal method to be implemented which does not perform caching
        """
        raise NotImplementedError

    def get_params(self, **tags):
        """
        Get the list of parameters, filtered by the provided tags.
        Some common tags include 'regularizable' and 'trainable'
        """
        tag_tuple = tuple(sorted(list(tags.items()), key=lambda x: x[0]))
        if tag_tuple not in self._cached_params:
            self._cached_params[tag_tuple] = self.get_params_internal(**tags)
        return self._cached_params[tag_tuple]

    def get_param_dtypes(self, **tags):
        tag_tuple = tuple(sorted(list(tags.items()), key=lambda x: x[0]))
        if tag_tuple not in self._cached_param_dtypes:
            params = self.get_params(**tags)
            param_values = tf.get_default_session().run(params)
            self._cached_param_dtypes[tag_tuple] = [
                val.dtype for val in param_values
            ]
        return self._cached_param_dtypes[tag_tuple]

    def get_param_shapes(self, **tags):
        tag_tuple = tuple(sorted(list(tags.items()), key=lambda x: x[0]))
        if tag_tuple not in self._cached_param_shapes:
            params = self.get_params(**tags)
            param_values = tf.get_default_session().run(params)
            self._cached_param_shapes[tag_tuple] = [
                val.shape for val in param_values
            ]
        return self._cached_param_shapes[tag_tuple]

    def get_param_values(self, **tags):
        params = self.get_params(**tags)
        param_values = tf.get_default_session().run(params)
        return flatten_tensors(param_values)

    def set_param_values(self, flattened_params, name=None, **tags):
        with tf.name_scope(name, 'set_param_values', [flattened_params]):
            debug = tags.pop('debug', False)
            param_values = unflatten_tensors(flattened_params,
                                             self.get_param_shapes(**tags))
            ops = []
            feed_dict = dict()
            for param, dtype, value in zip(
                    self.get_params(**tags), self.get_param_dtypes(**tags),
                    param_values):
                if param not in self._cached_assign_ops:
                    assign_placeholder = tf.placeholder(
                        dtype=param.dtype.base_dtype)
                    assign_op = tf.assign(param, assign_placeholder)
                    self._cached_assign_ops[param] = assign_op
                    self._cached_assign_placeholders[
                        param] = assign_placeholder
                ops.append(self._cached_assign_ops[param])
                feed_dict[self._cached_assign_placeholders[
                    param]] = value.astype(dtype)
                if debug:
                    print('setting value of %s' % param.name)
            tf.get_default_session().run(ops, feed_dict=feed_dict)

    def flat_to_params(self, flattened_params, **tags):
        return unflatten_tensors(flattened_params,
                                 self.get_param_shapes(**tags))

    def __getstate__(self):
        d = Serializable.__getstate__(self)
        global load_params
        if load_params:
            d['params'] = self.get_param_values()
        return d

    def __setstate__(self, d):
        Serializable.__setstate__(self, d)
        global load_params
        if load_params:
            tf.get_default_session().run(
                tf.variables_initializer(self.get_params()))
            self.set_param_values(d['params'])


class JointParameterized(Parameterized):
    def __init__(self, components):
        super(JointParameterized, self).__init__()
        self.components = components

    def get_params_internal(self, **tags):
        params = [
            param for comp in self.components
            for param in comp.get_params_internal(**tags)
        ]
        # only return unique parameters
        return sorted(set(params), key=hash)



import tensorflow as tf

from garage.tf.regressors import Regressor2
from tests.fixtures.models import SimpleMLPModel


class SimpleMLPRegressor(Regressor2):
    """Simple GaussianMLPRegressor for testing."""

    def __init__(self, input_shape, output_dim, name, *args, **kwargs):
        super().__init__(input_shape, output_dim, name)

        self.model = SimpleMLPModel(
            output_dim=self._output_dim, name='SimpleMLPModel')

        self._initialize()

    def _initialize(self):
        input_ph = tf.compat.v1.placeholder(
            tf.float32, shape=(None, ) + self._input_shape)
        with tf.compat.v1.variable_scope(self._name) as vs:
            self._variable_scope = vs
            self.model.build(input_ph)
        self.ys = None

    def fit(self, xs, ys):
        self.ys = ys

    def predict(self, xs):
        if self.ys is None:
            outputs = tf.compat.v1.get_default_session().run(
                self.model.networks['default'].outputs,
                feed_dict={self.model.networks['default'].input: xs})
            self.ys = outputs

        return self.ys

    def get_params_internal(self, *args, **kwargs):
        return self._variable_scope.trainable_variables()

    def __setstate__(self, state):
        """Object.__setstate__."""
        super().__setstate__(state)
        self._initialize()



import tensorflow as tf

from garage.core import Serializable
from garage.tf.core import LayersPowered
import garage.tf.core.layers as L
from garage.tf.core.layers import batch_norm
from garage.tf.misc import tensor_utils
from garage.tf.q_functions.base import QFunction


class ContinuousMLPQFunction(QFunction, LayersPowered, Serializable):
    """
    This class implements a q value network to predict q based on the input
    state and action. It uses an MLP to fit the function of Q(s, a).
    """

    def __init__(self,
                 env_spec,
                 name='ContinuousMLPQFunction',
                 hidden_sizes=(32, 32),
                 hidden_nonlinearity=tf.nn.relu,
                 action_merge_layer=-2,
                 output_nonlinearity=None,
                 input_include_goal=False,
                 bn=False):
        """
        Initialize class with multiple attributes.

        Args:
            env_spec():
            name(str, optional): A str contains the name of the policy.
            hidden_sizes(list or tuple, optional):
                A list of numbers of hidden units for all hidden layers.
            hidden_nonlinearity(optional):
                An activation shared by all fc layers.
            action_merge_layer(int, optional):
                An index to indicate when to merge action layer.
            output_nonlinearity(optional):
                An activation used by the output layer.
            bn(bool, optional):
                A bool to indicate whether normalize the layer or not.
        """
        Serializable.quick_init(self, locals())

        self.name = name
        self._env_spec = env_spec
        if input_include_goal:
            self._obs_dim = env_spec.observation_space.flat_dim_with_keys(
                ['observation', 'desired_goal'])
        else:
            self._obs_dim = env_spec.observation_space.flat_dim
        self._action_dim = env_spec.action_space.flat_dim
        self._hidden_sizes = hidden_sizes
        self._hidden_nonlinearity = hidden_nonlinearity
        self._action_merge_layer = action_merge_layer
        self._output_nonlinearity = output_nonlinearity
        self._batch_norm = bn
        self._f_qval, self._output_layer, self._obs_layer, self._action_layer = self.build_net(  # noqa: E501
            name=self.name)
        LayersPowered.__init__(self, [self._output_layer])

    def build_net(self, trainable=True, name=None):
        """
        Set up q network based on class attributes. This function uses layers
        defined in garage.tf.

        Args:
            reuse: A bool indicates whether reuse variables in the same scope.
            trainable: A bool indicates whether variables are trainable.
        """
        with tf.compat.v1.variable_scope(name):
            l_obs = L.InputLayer(shape=(None, self._obs_dim), name='obs')
            l_action = L.InputLayer(
                shape=(None, self._action_dim), name='actions')

            n_layers = len(self._hidden_sizes) + 1

            if n_layers > 1:
                action_merge_layer = \
                    (self._action_merge_layer % n_layers + n_layers) % n_layers
            else:
                action_merge_layer = 1

            l_hidden = l_obs

            for idx, size in enumerate(self._hidden_sizes):
                if self._batch_norm:
                    l_hidden = batch_norm(l_hidden)

                if idx == action_merge_layer:
                    l_hidden = L.ConcatLayer([l_hidden, l_action])

                l_hidden = L.DenseLayer(
                    l_hidden,
                    num_units=size,
                    nonlinearity=self._hidden_nonlinearity,
                    trainable=trainable,
                    name='hidden_%d' % (idx + 1))

            if action_merge_layer == n_layers:
                l_hidden = L.ConcatLayer([l_hidden, l_action])

            l_output = L.DenseLayer(
                l_hidden,
                num_units=1,
                nonlinearity=self._output_nonlinearity,
                trainable=trainable,
                name='output')

            output_var = L.get_output(l_output)

        f_qval = tensor_utils.compile_function(
            [l_obs.input_var, l_action.input_var], output_var)
        output_layer = l_output
        obs_layer = l_obs
        action_layer = l_action

        return f_qval, output_layer, obs_layer, action_layer

    def get_qval(self, observations, actions):
        return self._f_qval(observations, actions)

    def get_qval_sym(self, obs_var, action_var, name=None, **kwargs):
        with tf.name_scope(name, 'get_qval_sym', values=[obs_var, action_var]):
            qvals = L.get_output(self._output_layer, {
                self._obs_layer: obs_var,
                self._action_layer: action_var
            }, **kwargs)
            return qvals



from garage.np.algos import BatchPolopt


class DummyAlgo(BatchPolopt):
    """Dummy algo for test."""

    def __init__(self, policy, baseline):
        super().__init__(policy=policy,
                         baseline=baseline,
                         discount=0.1,
                         max_path_length=1,
                         n_samples=10)

    def train(self, runner):
        pass



import pickle

from dowel import logger
import numpy as np

from garage.experiment import deterministic
from garage.sampler.stateful_pool import SharedGlobal
from garage.sampler.stateful_pool import singleton_pool
from garage.sampler.utils import rollout


def _worker_init(g, id):
    if singleton_pool.n_parallel > 1:
        import os
        os.environ['CUDA_VISIBLE_DEVICES'] = ''
    g.worker_id = id


def initialize(n_parallel):
    singleton_pool.initialize(n_parallel)
    singleton_pool.run_each(
        _worker_init, [(id, ) for id in range(singleton_pool.n_parallel)])


def _get_scoped_g(g, scope):
    if scope is None:
        return g
    if not hasattr(g, 'scopes'):
        g.scopes = dict()
    if scope not in g.scopes:
        g.scopes[scope] = SharedGlobal()
        g.scopes[scope].worker_id = g.worker_id
    return g.scopes[scope]


def _worker_populate_task(g, env, policy, scope=None):
    g = _get_scoped_g(g, scope)
    g.env = pickle.loads(env)
    g.policy = pickle.loads(policy)


def _worker_terminate_task(g, scope=None):
    g = _get_scoped_g(g, scope)
    if getattr(g, 'env', None):
        g.env.close()
        g.env = None
    if getattr(g, 'policy', None):
        g.policy.terminate()
        g.policy = None


def populate_task(env, policy, scope=None):
    logger.log('Populating workers...')
    if singleton_pool.n_parallel > 1:
        singleton_pool.run_each(_worker_populate_task, [
            (pickle.dumps(env), pickle.dumps(policy), scope)
        ] * singleton_pool.n_parallel)
    else:
        # avoid unnecessary copying
        g = _get_scoped_g(singleton_pool.G, scope)
        g.env = env
        g.policy = policy
    logger.log('Populated')


def terminate_task(scope=None):
    singleton_pool.run_each(_worker_terminate_task,
                            [(scope, )] * singleton_pool.n_parallel)


def close():
    singleton_pool.close()


def _worker_set_seed(_, seed):
    logger.log('Setting seed to %d' % seed)
    deterministic.set_seed(seed)


def set_seed(seed):
    singleton_pool.run_each(_worker_set_seed,
                            [(seed + i, )
                             for i in range(singleton_pool.n_parallel)])


def _worker_set_policy_params(g, params, scope=None):
    g = _get_scoped_g(g, scope)
    g.policy.set_param_values(params)


def _worker_collect_one_path(g, max_path_length, scope=None):
    g = _get_scoped_g(g, scope)
    path = rollout(g.env, g.policy, max_path_length)
    return path, len(path['rewards'])


def sample_paths(policy_params,
                 max_samples,
                 max_path_length=np.inf,
                 scope=None):
    """
    :param policy_params: parameters for the policy. This will be updated on
     each worker process
    :param max_samples: desired maximum number of samples to be collected. The
     actual number of collected samples might be greater since all trajectories
     will be rolled out either until termination or until max_path_length is
     reached
    :param max_path_length: horizon / maximum length of a single trajectory
    :return: a list of collected paths
    """
    singleton_pool.run_each(
        _worker_set_policy_params,
        [(policy_params, scope)] * singleton_pool.n_parallel)
    return singleton_pool.run_collect(
        _worker_collect_one_path,
        threshold=max_samples,
        args=(max_path_length, scope),
        show_prog_bar=True)



#!/usr/bin/env python3

import argparse

import joblib
import tensorflow as tf

from garage.misc.console import query_yes_no
from garage.sampler.utils import rollout

if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('file', type=str, help='path to the snapshot file')
    parser.add_argument(
        '--max_path_length',
        type=int,
        default=1000,
        help='Max length of rollout')
    parser.add_argument('--speedup', type=float, default=1, help='Speedup')
    args = parser.parse_args()

    # If the snapshot file use tensorflow, do:
    # import tensorflow as tf
    # with tf.compat.v1.Session():
    #     [rest of the code]
    with tf.compat.v1.Session() as sess:
        data = joblib.load(args.file)
        policy = data['algo'].policy
        env = data['env']
        while True:
            path = rollout(
                env,
                policy,
                max_path_length=args.max_path_length,
                animated=True,
                speedup=args.speedup)
            if not query_yes_no('Continue simulation?'):
                break



"""Gaussian Conv Baseline."""
import numpy as np

from garage.core import Serializable
from garage.misc.overrides import overrides
from garage.np.baselines import Baseline
from garage.tf.core import Parameterized
from garage.tf.regressors import GaussianConvRegressor


class GaussianConvBaseline(Baseline, Parameterized, Serializable):
    """
    A Convolutional net Baseline with Gaussian dist output.

    Args:
        env_spec: a TfEnv.spec, which contains env spaces info.

    """

    def __init__(
            self,
            env_spec,
            subsample_factor=1.,
            regressor_args=None,
            name='GaussianConvBaseline',
    ):
        Parameterized.__init__(self)
        Serializable.quick_init(self, locals())
        super(GaussianConvBaseline, self).__init__(env_spec)

        if regressor_args is None:
            regressor_args = dict()

        self._regressor = GaussianConvRegressor(
            input_shape=env_spec.observation_space.shape,
            output_dim=1,
            name=name,
            **regressor_args)
        self.name = name

    @overrides
    def fit(self, paths):
        """Fit regressor based on paths."""
        observations = np.concatenate([p['observations'] for p in paths])
        returns = np.concatenate([p['returns'] for p in paths])
        self._regressor.fit(observations, returns.reshape((-1, 1)))

    @overrides
    def predict(self, path):
        """Predict value based on paths."""
        return self._regressor.predict(path['observations']).flatten()

    @overrides
    def get_param_values(self, **tags):
        """Get parameter values."""
        return self._regressor.get_param_values(**tags)

    @overrides
    def set_param_values(self, flattened_params, **tags):
        """Set parameter values to val."""
        self._regressor.set_param_values(flattened_params, **tags)

    @overrides
    def get_params_internal(self, **tags):
        """Get parameter values."""
        return self._regressor.get_params_internal(**tags)



import pickle
from unittest import mock

import numpy as np
import pytest
import tensorflow as tf

from garage.tf.optimizers import ConjugateGradientOptimizer, LbfgsOptimizer
from garage.tf.regressors import CategoricalMLPRegressorWithModel
from tests.fixtures import TfGraphTestCase


def get_labels(input_shape, xs):
    label = [0, 0]
    if input_shape == (1, ):
        ys = 0 if np.sin(xs) <= 0 else 1
        label[ys] = 1

    elif input_shape == (2, ):
        ys = int(np.round(xs[0])) ^ int(np.round(xs[1]))
        label[ys] = 1

    return label


def get_train_data(input_shape):
    if input_shape == (1, ):
        data = np.linspace(-np.pi, np.pi, 1000)
        obs = [{
            'observations': [[x]],
            'returns': [get_labels(input_shape, x)]
        } for x in data]

    elif input_shape == (2, ):
        data = [np.random.rand(2) for _ in range(1000)]
        obs = [{
            'observations': [x],
            'returns': [get_labels(input_shape, x)]
        } for x in data]
    return obs


def get_test_data(input_shape):
    if input_shape == (1, ):
        paths = {
            'observations': [[-np.pi / 2], [-np.pi / 3], [-np.pi / 4],
                             [np.pi / 4], [np.pi / 3], [np.pi / 4]]
        }
        expected = [[1, 0], [1, 0], [1, 0], [0, 1], [0, 1], [0, 1]]

    elif input_shape == (2, ):
        paths = {'observations': [[0, 0], [0, 1], [1, 0], [1, 1]]}
        expected = [[1, 0], [0, 1], [0, 1], [1, 0]]

    return paths, expected


class TestCategoricalMLPRegressorWithModel(TfGraphTestCase):
    @pytest.mark.parametrize('input_shape, output_dim', [((1, ), 2),
                                                         ((2, ), 2)])
    def test_fit_normalized(self, input_shape, output_dim):
        cmr = CategoricalMLPRegressorWithModel(
            input_shape=input_shape, output_dim=output_dim)
        obs = get_train_data(input_shape)

        observations = np.concatenate([p['observations'] for p in obs])
        returns = np.concatenate([p['returns'] for p in obs])
        returns = returns.reshape((-1, 2))

        for _ in range(150):
            cmr.fit(observations, returns)

        paths, expected = get_test_data(input_shape)

        prediction = cmr.predict(paths['observations'])

        assert np.allclose(prediction, expected, rtol=0, atol=0.1)

        x_mean = self.sess.run(cmr.model.networks['default'].x_mean)
        x_mean_expected = np.mean(observations, axis=0, keepdims=True)
        x_std = self.sess.run(cmr.model.networks['default'].x_std)
        x_std_expected = np.std(observations, axis=0, keepdims=True)

        assert np.allclose(x_mean, x_mean_expected)
        assert np.allclose(x_std, x_std_expected)

    @pytest.mark.parametrize('input_shape, output_dim', [((1, ), 2),
                                                         ((2, ), 2)])
    def test_fit_unnormalized(self, input_shape, output_dim):
        cmr = CategoricalMLPRegressorWithModel(
            input_shape=input_shape,
            output_dim=output_dim,
            normalize_inputs=False)
        obs = get_train_data(input_shape)

        observations = np.concatenate([p['observations'] for p in obs])
        returns = np.concatenate([p['returns'] for p in obs])
        returns = returns.reshape((-1, 2))

        for _ in range(150):
            cmr.fit(observations, returns)

        paths, expected = get_test_data(input_shape)

        prediction = cmr.predict(paths['observations'])

        assert np.allclose(prediction, expected, rtol=0, atol=0.1)

        x_mean = self.sess.run(cmr.model.networks['default'].x_mean)
        x_mean_expected = np.zeros_like(x_mean)
        x_std = self.sess.run(cmr.model.networks['default'].x_std)
        x_std_expected = np.ones_like(x_std)

        assert np.allclose(x_mean, x_mean_expected)
        assert np.allclose(x_std, x_std_expected)

    @pytest.mark.parametrize('input_shape, output_dim', [((1, ), 2),
                                                         ((2, ), 2)])
    def test_fit_without_initial_trust_region(self, input_shape, output_dim):
        cmr = CategoricalMLPRegressorWithModel(
            input_shape=input_shape,
            output_dim=output_dim,
            use_trust_region=False)
        obs = get_train_data(input_shape)

        observations = np.concatenate([p['observations'] for p in obs])
        returns = np.concatenate([p['returns'] for p in obs])
        returns = returns.reshape((-1, 2))

        for _ in range(150):
            cmr.fit(observations, returns)

        paths, expected = get_test_data(input_shape)

        prediction = cmr.predict(paths['observations'])

        assert np.allclose(prediction, expected, rtol=0, atol=0.1)

        x_mean = self.sess.run(cmr.model.networks['default'].x_mean)
        x_mean_expected = np.mean(observations, axis=0, keepdims=True)
        x_std = self.sess.run(cmr.model.networks['default'].x_std)
        x_std_expected = np.std(observations, axis=0, keepdims=True)

        assert np.allclose(x_mean, x_mean_expected)
        assert np.allclose(x_std, x_std_expected)

    @pytest.mark.parametrize('output_dim, input_shape',
                             [(1, (1, )), (1, (2, )), (2, (3, )), (2, (1, 1)),
                              (3, (2, 2))])
    def test_dist_info_sym(self, output_dim, input_shape):
        cmr = CategoricalMLPRegressorWithModel(
            input_shape=input_shape, output_dim=output_dim)

        new_xs_var = tf.compat.v1.placeholder(
            tf.float32, shape=(None, ) + input_shape)

        data = np.random.random(size=input_shape)
        label = np.random.randint(0, output_dim)
        one_hot_label = np.zeros(output_dim)
        one_hot_label[label] = 1

        outputs = cmr.dist_info_sym(new_xs_var, name='dist_info_sym')
        prob = self.sess.run(outputs, feed_dict={new_xs_var: [data]})

        expected_prob = cmr._f_prob(np.asarray([data]))

        assert np.allclose(prob['prob'], expected_prob, rtol=0, atol=1e-5)

    @pytest.mark.parametrize('output_dim, input_shape',
                             [(1, (1, )), (1, (2, )), (2, (3, )), (2, (1, 1)),
                              (3, (2, 2))])
    def test_log_likelihood_sym(self, output_dim, input_shape):
        cmr = CategoricalMLPRegressorWithModel(
            input_shape=input_shape, output_dim=output_dim)

        new_xs_var = tf.compat.v1.placeholder(
            tf.float32, shape=(None, ) + input_shape)
        new_ys_var = tf.compat.v1.placeholder(
            dtype=tf.float32, name='ys', shape=(None, output_dim))

        data = np.random.random(size=input_shape)
        label = np.random.randint(0, output_dim)
        one_hot_label = np.zeros(output_dim)
        one_hot_label[label] = 1

        ll = cmr.predict_log_likelihood([data], [one_hot_label])

        outputs = cmr.log_likelihood_sym(new_xs_var, new_ys_var, name='ll_sym')

        ll_from_sym = self.sess.run(
            outputs,
            feed_dict={
                new_xs_var: [data],
                new_ys_var: [one_hot_label]
            })

        assert np.allclose(ll, ll_from_sym, rtol=0, atol=1e-5)

    @mock.patch('tests.garage.tf.regressors.'
                'test_categorical_mlp_regressor_with_model.'
                'LbfgsOptimizer')
    @mock.patch('tests.garage.tf.regressors.'
                'test_categorical_mlp_regressor_with_model.'
                'ConjugateGradientOptimizer')
    def test_optimizer_args(self, mock_cg, mock_lbfgs):
        lbfgs_args = dict(max_opt_itr=25)
        cg_args = dict(cg_iters=15)
        cmr = CategoricalMLPRegressorWithModel(
            input_shape=(1, ),
            output_dim=2,
            optimizer=LbfgsOptimizer,
            optimizer_args=lbfgs_args,
            tr_optimizer=ConjugateGradientOptimizer,
            tr_optimizer_args=cg_args,
            use_trust_region=True)

        assert mock_lbfgs.return_value is cmr._optimizer
        assert mock_cg.return_value is cmr._tr_optimizer

        mock_lbfgs.assert_called_with(max_opt_itr=25)
        mock_cg.assert_called_with(cg_iters=15)

    def test_is_pickleable(self):
        cmr = CategoricalMLPRegressorWithModel(input_shape=(1, ), output_dim=2)

        with tf.compat.v1.variable_scope(
                'CategoricalMLPRegressorWithModel/NormalizedInputMLPModel',
                reuse=True):
            bias = tf.compat.v1.get_variable('mlp/hidden_0/bias')
        bias.load(tf.ones_like(bias).eval())

        result1 = cmr.predict(np.ones((1, 1)))

        h = pickle.dumps(cmr)
        with tf.compat.v1.Session(graph=tf.Graph()):
            cmr_pickled = pickle.loads(h)
            result2 = cmr_pickled.predict(np.ones((1, 1)))
            assert np.array_equal(result1, result2)

    def test_is_pickleable2(self):
        cmr = CategoricalMLPRegressorWithModel(input_shape=(1, ), output_dim=2)

        with tf.compat.v1.variable_scope(
                'CategoricalMLPRegressorWithModel/NormalizedInputMLPModel',
                reuse=True):
            x_mean = tf.compat.v1.get_variable('normalized_vars/x_mean')
        x_mean.load(tf.ones_like(x_mean).eval())
        x1 = cmr.model.networks['default'].x_mean.eval()
        h = pickle.dumps(cmr)
        with tf.compat.v1.Session(graph=tf.Graph()):
            cmr_pickled = pickle.loads(h)
            x2 = cmr_pickled.model.networks['default'].x_mean.eval()
            assert np.array_equal(x1, x2)



"""BatchSampler which uses VecEnvExecutor to run multiple environments."""
import itertools
import pickle

from dowel import logger, tabular
import numpy as np

from garage.misc import tensor_utils
from garage.misc.overrides import overrides
from garage.misc.prog_bar_counter import ProgBarCounter
from garage.sampler.batch_sampler import BatchSampler
from garage.sampler.stateful_pool import singleton_pool
from garage.sampler.utils import truncate_paths
from garage.sampler.vec_env_executor import VecEnvExecutor


class OnPolicyVectorizedSampler(BatchSampler):
    """BatchSampler which uses VecEnvExecutor to run multiple environments."""

    def __init__(self, algo, env, n_envs=None):
        if n_envs is None:
            n_envs = singleton_pool.n_parallel * 4
        super().__init__(algo, env)
        self.n_envs = n_envs

    @overrides
    def start_worker(self):
        """Start workers."""
        n_envs = self.n_envs
        envs = [pickle.loads(pickle.dumps(self.env)) for _ in range(n_envs)]
        self.vec_env = VecEnvExecutor(
            envs=envs, max_path_length=self.algo.max_path_length)
        self.env_spec = self.env.spec

    @overrides
    def shutdown_worker(self):
        """Shutdown workers."""
        self.vec_env.close()

    @overrides
    def obtain_samples(self, itr, batch_size=None, whole_paths=True):
        """Obtain samples."""
        logger.log('Obtaining samples for iteration %d...' % itr)

        if not batch_size:
            batch_size = self.algo.max_path_length * self.n_envs

        paths = []
        n_samples = 0
        obses = self.vec_env.reset()
        dones = np.asarray([True] * self.vec_env.num_envs)
        running_paths = [None] * self.vec_env.num_envs

        pbar = ProgBarCounter(batch_size)
        policy_time = 0
        env_time = 0
        process_time = 0

        policy = self.algo.policy

        import time
        while n_samples < batch_size:
            t = time.time()
            policy.reset(dones)

            actions, agent_infos = policy.get_actions(obses)

            policy_time += time.time() - t
            t = time.time()
            next_obses, rewards, dones, env_infos = self.vec_env.step(actions)
            env_time += time.time() - t
            t = time.time()

            agent_infos = tensor_utils.split_tensor_dict_list(agent_infos)
            env_infos = tensor_utils.split_tensor_dict_list(env_infos)
            if env_infos is None:
                env_infos = [dict() for _ in range(self.vec_env.num_envs)]
            if agent_infos is None:
                agent_infos = [dict() for _ in range(self.vec_env.num_envs)]
            for idx, observation, action, reward, env_info, agent_info, done in zip(  # noqa: E501
                    itertools.count(), obses, actions, rewards, env_infos,
                    agent_infos, dones):
                if running_paths[idx] is None:
                    running_paths[idx] = dict(
                        observations=[],
                        actions=[],
                        rewards=[],
                        env_infos=[],
                        agent_infos=[],
                    )
                running_paths[idx]['observations'].append(observation)
                running_paths[idx]['actions'].append(action)
                running_paths[idx]['rewards'].append(reward)
                running_paths[idx]['env_infos'].append(env_info)
                running_paths[idx]['agent_infos'].append(agent_info)
                if done:
                    paths.append(
                        dict(observations=self.env_spec.observation_space.
                             flatten_n(running_paths[idx]['observations']),
                             actions=self.env_spec.action_space.flatten_n(
                                 running_paths[idx]['actions']),
                             rewards=tensor_utils.stack_tensor_list(
                                 running_paths[idx]['rewards']),
                             env_infos=tensor_utils.stack_tensor_dict_list(
                                 running_paths[idx]['env_infos']),
                             agent_infos=tensor_utils.stack_tensor_dict_list(
                                 running_paths[idx]['agent_infos'])))
                    n_samples += len(running_paths[idx]['rewards'])
                    running_paths[idx] = None

            process_time += time.time() - t
            pbar.inc(len(obses))
            obses = next_obses

        pbar.stop()

        tabular.record('PolicyExecTime', policy_time)
        tabular.record('EnvExecTime', env_time)
        tabular.record('ProcessExecTime', process_time)

        if whole_paths:
            return paths
        else:
            paths_truncated = truncate_paths(paths, batch_size)
            return paths_truncated



"""Regressors for TensorFlow-based algorithms."""
from garage.tf.regressors.base import StochasticRegressor
from garage.tf.regressors.base2 import Regressor2
from garage.tf.regressors.base2 import StochasticRegressor2
from garage.tf.regressors.bernoulli_mlp_regressor import BernoulliMLPRegressor
from garage.tf.regressors.bernoulli_mlp_regressor_with_model import (
    BernoulliMLPRegressorWithModel)
from garage.tf.regressors.categorical_mlp_regressor import (
    CategoricalMLPRegressor)
from garage.tf.regressors.categorical_mlp_regressor_with_model import (
    CategoricalMLPRegressorWithModel)
from garage.tf.regressors.continuous_mlp_regressor import (
    ContinuousMLPRegressor)
from garage.tf.regressors.continuous_mlp_regressor_with_model import (
    ContinuousMLPRegressorWithModel)
from garage.tf.regressors.gaussian_cnn_regressor_model import (
    GaussianCNNRegressorModel)
from garage.tf.regressors.gaussian_cnn_regressor_with_model import (
    GaussianCNNRegressorWithModel)
from garage.tf.regressors.gaussian_conv_regressor import GaussianConvRegressor
from garage.tf.regressors.gaussian_mlp_regressor import GaussianMLPRegressor

__all__ = [
    'BernoulliMLPRegressor', 'BernoulliMLPRegressorWithModel',
    'CategoricalMLPRegressor', 'CategoricalMLPRegressorWithModel',
    'ContinuousMLPRegressor', 'ContinuousMLPRegressorWithModel',
    'GaussianCNNRegressorModel', 'GaussianCNNRegressorWithModel',
    'GaussianMLPRegressor', 'GaussianConvRegressor', 'Regressor2',
    'StochasticRegressor', 'StochasticRegressor2'
]



import gym
import numpy as np

from garage.envs import EnvSpec
from tests.fixtures.envs.dummy import DummyEnv


class DummyDictEnv(DummyEnv):
    """A dummy dict environment."""

    def __init__(self, random=True):
        super().__init__(random)
        self.spec = EnvSpec(
            action_space=self.action_space,
            observation_space=self.observation_space)

    @property
    def observation_space(self):
        """Return an observation space."""

        return gym.spaces.Dict({
            'achieved_goal':
            gym.spaces.Box(
                low=-200., high=200., shape=(3, ), dtype=np.float32),
            'desired_goal':
            gym.spaces.Box(
                low=-200., high=200., shape=(3, ), dtype=np.float32),
            'observation':
            gym.spaces.Box(
                low=-200., high=200., shape=(25, ), dtype=np.float32)
        })

    @property
    def action_space(self):
        """Return an action space."""
        return gym.spaces.Dict({
            'action':
            gym.spaces.Box(low=-5.0, high=5.0, shape=(1, ), dtype=np.float32)
        })

    def reset(self):
        """Reset the environment."""
        return self.observation_space.sample()

    def step(self, action):
        """Step the environment."""
        return self.observation_space.sample(), 0, True, dict()



import akro
import numpy as np
import tensorflow as tf

from garage.core import Serializable
from garage.misc.overrides import overrides
from garage.tf.core import LayersPowered
import garage.tf.core.layers as L
from garage.tf.core.network import LSTMNetwork
from garage.tf.distributions import RecurrentCategorical
from garage.tf.misc import tensor_utils
from garage.tf.policies.base import StochasticPolicy


class CategoricalLSTMPolicy(StochasticPolicy, LayersPowered, Serializable):
    def __init__(self,
                 env_spec,
                 name='CategoricalLSTMPolicy',
                 hidden_dim=32,
                 hidden_nonlinearity=tf.nn.tanh,
                 recurrent_nonlinearity=tf.nn.sigmoid,
                 recurrent_w_x_init=L.XavierUniformInitializer(),
                 recurrent_w_h_init=L.OrthogonalInitializer(),
                 output_nonlinearity=tf.nn.softmax,
                 output_w_init=L.XavierUniformInitializer(),
                 feature_network=None,
                 prob_network=None,
                 state_include_action=True,
                 forget_bias=1.0,
                 use_peepholes=False,
                 lstm_layer_cls=L.LSTMLayer):
        """
        :param env_spec: A spec for the env.
        :param hidden_dim: dimension of hidden layer
        :param hidden_nonlinearity: nonlinearity used for each hidden layer
        :return:
        """
        assert isinstance(env_spec.action_space, akro.Discrete)

        self._prob_network_name = 'prob_network'
        with tf.compat.v1.variable_scope(name, 'CategoricalLSTMPolicy'):
            Serializable.quick_init(self, locals())
            super(CategoricalLSTMPolicy, self).__init__(env_spec)

            obs_dim = env_spec.observation_space.flat_dim
            action_dim = env_spec.action_space.flat_dim

            if state_include_action:
                input_dim = obs_dim + action_dim
            else:
                input_dim = obs_dim

            l_input = L.InputLayer(shape=(None, None, input_dim), name='input')

            if feature_network is None:
                feature_dim = input_dim
                l_flat_feature = None
                l_feature = l_input
            else:
                feature_dim = feature_network.output_layer.output_shape[-1]
                l_flat_feature = feature_network.output_layer
                l_feature = L.OpLayer(
                    l_flat_feature,
                    extras=[l_input],
                    name='reshape_feature',
                    op=lambda flat_feature, input: tf.reshape(
                        flat_feature,
                        tf.stack([
                            tf.shape(input)[0],
                            tf.shape(input)[1], feature_dim
                        ])),
                    shape_op=lambda _, input_shape: (input_shape[
                        0], input_shape[1], feature_dim))

            if prob_network is None:
                prob_network = LSTMNetwork(
                    input_shape=(feature_dim, ),
                    input_layer=l_feature,
                    output_dim=env_spec.action_space.n,
                    hidden_dim=hidden_dim,
                    hidden_nonlinearity=hidden_nonlinearity,
                    recurrent_nonlinearity=recurrent_nonlinearity,
                    recurrent_w_x_init=recurrent_w_x_init,
                    recurrent_w_h_init=recurrent_w_h_init,
                    output_nonlinearity=output_nonlinearity,
                    output_w_init=output_w_init,
                    forget_bias=forget_bias,
                    use_peepholes=use_peepholes,
                    lstm_layer_cls=lstm_layer_cls,
                    name=self._prob_network_name)

            self.prob_network = prob_network
            self.feature_network = feature_network
            self.l_input = l_input
            self.state_include_action = state_include_action

            flat_input_var = tf.compat.v1.placeholder(
                dtype=tf.float32, shape=(None, input_dim), name='flat_input')
            if feature_network is None:
                feature_var = flat_input_var
            else:
                with tf.name_scope('feature_network', values=[flat_input_var]):
                    feature_var = L.get_output(
                        l_flat_feature,
                        {feature_network.input_layer: flat_input_var})

            with tf.name_scope(self._prob_network_name, values=[feature_var]):
                out_prob_step, out_prob_hidden, out_step_cell = L.get_output(
                    [
                        prob_network.step_output_layer,
                        prob_network.step_hidden_layer,
                        prob_network.step_cell_layer
                    ], {prob_network.step_input_layer: feature_var})

            self.f_step_prob = tensor_utils.compile_function([
                flat_input_var,
                prob_network.step_prev_state_layer.input_var,
            ], [out_prob_step, out_prob_hidden, out_step_cell])

            self.input_dim = input_dim
            self.action_dim = action_dim
            self.hidden_dim = hidden_dim
            self.name = name

            self.prev_actions = None
            self.prev_hiddens = None
            self.prev_cells = None
            self.dist = RecurrentCategorical(env_spec.action_space.n)

            out_layers = [prob_network.output_layer]
            if feature_network is not None:
                out_layers.append(feature_network.output_layer)

            LayersPowered.__init__(self, out_layers)

    @overrides
    def dist_info_sym(self, obs_var, state_info_vars, name=None):
        with tf.name_scope(name, 'dist_info_sym', [obs_var, state_info_vars]):
            n_batches = tf.shape(obs_var)[0]
            n_steps = tf.shape(obs_var)[1]
            obs_var = tf.reshape(obs_var, tf.stack([n_batches, n_steps, -1]))
            obs_var = tf.cast(obs_var, tf.float32)
            if self.state_include_action:
                prev_action_var = state_info_vars['prev_action']
                prev_action_var = tf.cast(prev_action_var, tf.float32)
                all_input_var = tf.concat(
                    axis=2, values=[obs_var, prev_action_var])
            else:
                all_input_var = obs_var
            if self.feature_network is None:
                with tf.name_scope(
                        self._prob_network_name, values=[all_input_var]):
                    prob = L.get_output(self.prob_network.output_layer,
                                        {self.l_input: all_input_var})
                return dict(prob=prob)
            else:
                flat_input_var = tf.reshape(all_input_var,
                                            (-1, self.input_dim))
                with tf.name_scope(
                        self._prob_network_name,
                        values=[all_input_var, flat_input_var]):
                    prob = L.get_output(
                        self.prob_network.output_layer, {
                            self.l_input: all_input_var,
                            self.feature_network.input_layer: flat_input_var
                        })
                return dict(prob=prob)

    @property
    def vectorized(self):
        return True

    def reset(self, dones=None):
        if dones is None:
            dones = [True]
        dones = np.asarray(dones)
        if self.prev_actions is None or len(dones) != len(self.prev_actions):
            self.prev_actions = np.zeros((len(dones),
                                          self.action_space.flat_dim))
            self.prev_hiddens = np.zeros((len(dones), self.hidden_dim))
            self.prev_cells = np.zeros((len(dones), self.hidden_dim))

        self.prev_actions[dones] = 0.
        self.prev_hiddens[dones] = self.prob_network.hid_init_param.eval()
        self.prev_cells[dones] = self.prob_network.cell_init_param.eval()

    @overrides
    def get_action(self, observation):
        actions, agent_infos = self.get_actions([observation])
        return actions[0], {k: v[0] for k, v in agent_infos.items()}

    @overrides
    def get_actions(self, observations):
        flat_obs = self.observation_space.flatten_n(observations)
        if self.state_include_action:
            assert self.prev_actions is not None
            all_input = np.concatenate([flat_obs, self.prev_actions], axis=-1)
        else:
            all_input = flat_obs
        probs, hidden_vec, cell_vec = self.f_step_prob(
            all_input,
            np.concatenate([self.prev_hiddens, self.prev_cells], axis=-1))
        actions = list(map(self.action_space.weighted_sample, probs))
        prev_actions = self.prev_actions
        self.prev_actions = self.action_space.flatten_n(actions)
        self.prev_hiddens = hidden_vec
        self.prev_cells = cell_vec
        agent_info = dict(prob=probs)
        if self.state_include_action:
            agent_info['prev_action'] = np.copy(prev_actions)
        return actions, agent_info

    @property
    @overrides
    def recurrent(self):
        return True

    @property
    def distribution(self):
        return self.dist

    @property
    def state_info_specs(self):
        if self.state_include_action:
            return [
                ('prev_action', (self.action_dim, )),
            ]
        else:
            return []



import akro
import tensorflow as tf

from garage.core import Serializable
from garage.misc.overrides import overrides
from garage.tf.core import LayersPowered
import garage.tf.core.layers as L
from garage.tf.core.network import MLP
from garage.tf.distributions import Categorical
from garage.tf.misc import tensor_utils
from garage.tf.policies.base import StochasticPolicy


class CategoricalMLPPolicy(StochasticPolicy, LayersPowered, Serializable):
    def __init__(
            self,
            env_spec,
            name='CategoricalMLPPolicy',
            hidden_sizes=(32, 32),
            hidden_nonlinearity=tf.nn.tanh,
            prob_network=None,
    ):
        """
        CategoricalMLPPolicy.

        A policy that uses a MLP to estimate a categorical distribution.

        Args:
            env_spec (garage.envs.env_spec.EnvSpec): Environment specification.
            hidden_sizes (list[int]): Output dimension of dense layer(s).
                For example, (32, 32) means the MLP of this policy consists
                of two hidden layers, each with 32 hidden units.
            hidden_nonlinearity: Activation function for
                intermediate dense layer(s).
            prob_network (tf.Tensor): manually specified network for this
                policy. If None, a MLP with the network parameters will be
                created. If not None, other network params are ignored.
        """
        assert isinstance(env_spec.action_space, akro.Discrete)

        Serializable.quick_init(self, locals())

        self.name = name
        self._prob_network_name = 'prob_network'
        with tf.compat.v1.variable_scope(name, 'CategoricalMLPPolicy'):
            if prob_network is None:
                prob_network = MLP(
                    input_shape=(env_spec.observation_space.flat_dim, ),
                    output_dim=env_spec.action_space.n,
                    hidden_sizes=hidden_sizes,
                    hidden_nonlinearity=hidden_nonlinearity,
                    output_nonlinearity=tf.nn.softmax,
                    name=self._prob_network_name,
                )

            self._l_prob = prob_network.output_layer
            self._l_obs = prob_network.input_layer
            with tf.name_scope(self._prob_network_name):
                prob_network_outputs = L.get_output(prob_network.output_layer)
            self._f_prob = tensor_utils.compile_function(
                [prob_network.input_layer.input_var], prob_network_outputs)

            self._dist = Categorical(env_spec.action_space.n)

            super(CategoricalMLPPolicy, self).__init__(env_spec)
            LayersPowered.__init__(self, [prob_network.output_layer])

    @property
    def vectorized(self):
        return True

    @overrides
    def dist_info_sym(self, obs_var, state_info_vars=None, name=None):
        with tf.name_scope(name, 'dist_info_sym', [obs_var, state_info_vars]):
            with tf.name_scope(self._prob_network_name, values=[obs_var]):
                prob = L.get_output(
                    self._l_prob, {self._l_obs: tf.cast(obs_var, tf.float32)})
            return dict(prob=prob)

    @overrides
    def dist_info(self, obs, state_infos=None):
        return dict(prob=self._f_prob(obs))

    # The return value is a pair. The first item is a matrix (N, A), where each
    # entry corresponds to the action value taken. The second item is a vector
    # of length N, where each entry is the density value for that action, under
    # the current policy
    @overrides
    def get_action(self, observation):
        flat_obs = self.observation_space.flatten(observation)
        prob = self._f_prob([flat_obs])[0]
        action = self.action_space.weighted_sample(prob)
        return action, dict(prob=prob)

    def get_actions(self, observations):
        flat_obs = self.observation_space.flatten_n(observations)
        probs = self._f_prob(flat_obs)
        actions = list(map(self.action_space.weighted_sample, probs))
        return actions, dict(prob=probs)

    @property
    def distribution(self):
        return self._dist



import signal
import time

import numpy as np

from garage.misc import tensor_utils


def rollout(env,
            agent,
            max_path_length=np.inf,
            animated=False,
            speedup=1,
            always_return_paths=False):
    observations = []
    actions = []
    rewards = []
    agent_infos = []
    env_infos = []
    o = env.reset()
    agent.reset()
    path_length = 0
    if animated:
        env.render()
    while path_length < max_path_length:
        a, agent_info = agent.get_action(o)
        next_o, r, d, env_info = env.step(a)
        observations.append(env.observation_space.flatten(o))
        rewards.append(r)
        actions.append(env.action_space.flatten(a))
        agent_infos.append(agent_info)
        env_infos.append(env_info)
        path_length += 1
        if d:
            break
        o = next_o
        if animated:
            env.render()
            timestep = 0.05
            time.sleep(timestep / speedup)
    if animated and not always_return_paths:
        return None

    return dict(
        observations=tensor_utils.stack_tensor_list(observations),
        actions=tensor_utils.stack_tensor_list(actions),
        rewards=tensor_utils.stack_tensor_list(rewards),
        agent_infos=tensor_utils.stack_tensor_dict_list(agent_infos),
        env_infos=tensor_utils.stack_tensor_dict_list(env_infos),
    )


def truncate_paths(paths, max_samples):
    """
    Truncate the list of paths so that the total number of samples is exactly
    equal to max_samples. This is done by removing extra paths at the end of
    the list, and make the last path shorter if necessary
    :param paths: a list of paths
    :param max_samples: the absolute maximum number of samples
    :return: a list of paths, truncated so that the number of samples adds up
    to max-samples
    """
    # chop samples collected by extra paths
    # make a copy
    paths = list(paths)
    total_n_samples = sum(len(path['rewards']) for path in paths)
    while paths and total_n_samples - len(paths[-1]['rewards']) >= max_samples:
        total_n_samples -= len(paths.pop(-1)['rewards'])
    if paths:
        last_path = paths.pop(-1)
        truncated_last_path = dict()
        truncated_len = len(
            last_path['rewards']) - (total_n_samples - max_samples)
        for k, v in last_path.items():
            if k in ['observations', 'actions', 'rewards']:
                truncated_last_path[k] = tensor_utils.truncate_tensor_list(
                    v, truncated_len)
            elif k in ['env_infos', 'agent_infos']:
                truncated_last_path[k] = tensor_utils.truncate_tensor_dict(
                    v, truncated_len)
            else:
                raise NotImplementedError
        paths.append(truncated_last_path)
    return paths


def center_advantages(advantages):
    return (advantages - np.mean(advantages)) / (advantages.std() + 1e-8)


def shift_advantages_to_positive(advantages):
    return (advantages - np.min(advantages)) + 1e-8


def sign(x):
    return 1. * (x >= 0) - 1. * (x < 0)


class MaskSignals():
    """Context Manager to mask a list of signals."""

    def __init__(self, signals):
        self.signals = signals

    def __enter__(self):
        signal.pthread_sigmask(signal.SIG_BLOCK, self.signals)

    def __exit__(self, *args):
        signal.pthread_sigmask(signal.SIG_UNBLOCK, self.signals)


mask_signals = MaskSignals



"""Dummy Policy for algo tests.."""
import numpy as np

from garage.core import Serializable
from garage.misc.overrides import overrides
from garage.np.policies import Policy
from tests.fixtures.distributions import DummyDistribution


class DummyPolicy(Policy, Serializable):
    """Dummy Policy."""

    def __init__(
            self,
            env_spec,
    ):
        Serializable.quick_init(self, locals())
        super().__init__(env_spec=env_spec)

    @overrides
    def get_action(self, observation):
        """Return action."""
        return self.action_space.sample(), dict()

    def get_actions(self, obses):
        """Return actions."""
        n = len(obses)
        action, action_info = self.get_action(None)
        return [action] * n, action_info

    @overrides
    def get_params_internal(self, **tags):
        """Return a list of policy internal params."""
        return []

    @overrides
    def get_param_values(self, **tags):
        """Return values of params."""
        return np.random.uniform(-1, 1, 1000)

    @overrides
    def reset(self, *args, **kwargs):
        pass

    @property
    def distribution(self):
        """Return the distribution."""
        return DummyDistribution()

    @property
    def vectorized(self):
        return True


class DummyPolicyWithoutVectorized(DummyPolicy):
    """Dummy Policy without vectorized"""

    def __init__(self, env_spec):
        super().__init__(env_spec=env_spec)

    @property
    def vectorized(self):
        return False



"""Baseline estimators for TensorFlow-based algorithms."""
from garage.tf.baselines.continuous_mlp_baseline import ContinuousMLPBaseline
from garage.tf.baselines.continuous_mlp_baseline_with_model import (
    ContinuousMLPBaselineWithModel)
from garage.tf.baselines.gaussian_cnn_baseline_with_model import (
    GaussianCNNBaselineWithModel)
from garage.tf.baselines.gaussian_conv_baseline import GaussianConvBaseline
from garage.tf.baselines.gaussian_mlp_baseline import GaussianMLPBaseline

__all__ = [
    'ContinuousMLPBaseline',
    'ContinuousMLPBaselineWithModel',
    'GaussianConvBaseline',
    'GaussianCNNBaselineWithModel',
    'GaussianMLPBaseline',
]
